//  File name: $HeadURL: https://www.rheinahrcampus.de/svn/repos/predict4a/pLib/src/H2OMapping.cpp $
//  Revision: $Revision: 53 $
//  Last modified: $Date: 2012-05-02 10:46:15 +0200 (Mi, 02 Mai 2012) $
//  Last modified by: $Author: neeb $

/*
 * H2OMapping.cpp
 *
 *  Created on: 03.02.2012
 *      Author: neeb

	Copyright 2012 Heiko Neeb

    This file is part of PREDICT-MS (Professional Remagen Diagnostic Image Calculation Tool for MS).

    PREDICT-MS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PREDICT-MS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with "predict".  If not, see <http://www.gnu.org/licenses/>.

	PREDICT-MS is designed to be used for scientific purposes only. Do NOT
	base any diagnosis on the results obtained with this program

	The software is based on the following work which the authors kindly ask to reference in
	any scientific work which is based on PREDICT-MS or parts of it:

	V. Tonkova, V. Arhelger, J. Schenk, H. Neeb. Rapid Myelin Water Content Mapping on Clinical MR Systems.
	Z Med Phys (2012): http://dx.doi.org/10.1016/j.zemedi.2011.09.005

	H. Neeb, V. Ermer, T. Stocker, N.J. Shah. Fast quantitative mapping of absolute water content
	with full brain coverage. NeuroImage (2008). 42(3): 10941109

 */

#include "H2OMapping.h"
#include "qImageMap.h"
#include "qImageUtils.h"
#include "qImageProcessing.h"
#include <math.h>

qH2OMapping::qH2OMapping() {
	applyHighFieldCor = false;
	coilCorCalculated = false;
}

qH2OMapping::~qH2OMapping() {
	// TODO Auto-generated destructor stub
}


void qH2OMapping::calcCoilCor(){

	std::list<boost::shared_ptr<qImage> >::iterator maskIter;
	std::list<boost::shared_ptr<qImage> >::iterator epiBodyIter;
	std::list<boost::shared_ptr<qImage> >::iterator epiHeadIter;
	boost::shared_ptr<qImage> curImg;

	// get size of low resolution (EPI) images
	int nColLowRes = (*EPIBody.begin())->getNcols();
	int nRowLowRes = (*EPIBody.begin())->getNrows();

	// create new (temporary) low resolution image
	boost::shared_ptr<qImage> CC = boost::shared_ptr<qImageMap>(new qImageMap(nRowLowRes, nColLowRes));
	matrix<int>& CCMat = CC->getPixelMatrix();

	// Loop over all points and calculate ratio of EPI body and EPI head
	for (int slice = 1; slice <= nSlices; slice++) {

		// set iterators properly (required for parallel processing)
		epiBodyIter = EPIBody.begin();
		std::advance(epiBodyIter, slice - 1);
		epiHeadIter = EPIHead.begin();
		std::advance(epiHeadIter, slice - 1);
		maskIter = MaskLowRes.begin();
		std::advance(maskIter, slice - 1);

		matrix<int>& epiBodyMat = (*epiBodyIter)->getPixelMatrix();
		matrix<int>& epiHeadMat = (*epiHeadIter)->getPixelMatrix();
		matrix<int>& maskMat = (*maskIter)->getPixelMatrix();

		/* loop over all points in matrix */
		for (int row = 0; row < nRowLowRes; row++) {
			for (int col = 0; col < nColLowRes; col++) {
				if( maskMat(row,col)>0 && epiHeadMat(row,col)>0 ) {
					CCMat(row,col)=(int)(1000.0*epiBodyMat(row,col)/(double)epiHeadMat(row,col));
				}
				else
					CCMat(row,col)=0;
			}
		}

		// filter image with median filter
		FilterKernel flipMapKern; //dummy
		flipMapKern.createMeanKernel(3,3,1); // ... we only need the size of the filter mask
		qImageProcessing::filterWithKernel(CC,flipMapKern, qImageProcessing::OMIT_ZEROS, qImageProcessing::MEDIAN_FILTER);

		// Re-size low-res image
		curImg = qImageUtils::resize<qImageMap>(nCol, nRow, CC);
        CoilCor.push_back(curImg);
	}

	// filter stack again with 3D mean filter
	FilterKernel flipMapKern;
	flipMapKern.createMeanKernel(3,3,5); // ... we only need the size of the filter mask for the median filter
	qImageProcessing::filterStackWithKernel(CoilCor,flipMapKern, qImageProcessing::OMIT_ZEROS, qImageProcessing::MEDIAN_FILTER);

	flipMapKern.createMeanKernel(3,3,3);
	qImageProcessing::filterStackWithKernel(CoilCor,flipMapKern, qImageProcessing::OMIT_ZEROS, qImageProcessing::KERNEL_FILTER);

	flipMapKern.createMeanKernel(5,5,1);
	qImageProcessing::filterStackWithKernel(CoilCor,flipMapKern, qImageProcessing::FILL_ZEROS_ONLY, qImageProcessing::KERNEL_FILTER);

	coilCorCalculated = true;
}



void qH2OMapping::calculateHighFieldCorrection(
		std::list<boost::shared_ptr<qImage> > WMMask,
		std::list<boost::shared_ptr<qImage> > GMMask,
		std::list<boost::shared_ptr<qImage> > CSFMask,
		double medianWM,
		double medianGM,
		double medianCSF) {

	std::list<boost::shared_ptr<qImage> >::iterator H2OIter;
	std::list<boost::shared_ptr<qImage> >::iterator WMIter;
	std::list<boost::shared_ptr<qImage> >::iterator GMIter;
	std::list<boost::shared_ptr<qImage> >::iterator CSFIter;
	std::list<boost::shared_ptr<qImage> > H2OWM;
	std::list<boost::shared_ptr<qImage> > H2OGM;
	std::list<boost::shared_ptr<qImage> > H2OCSF;
	boost::shared_ptr<qImage> curImg;

	std::vector<double> medVecWM;
	std::vector<double> medVecGM;
	std::vector<double> medVecCSF;

	// delete any possibly existing list entries first
	if ( !HighBCor.empty() ) {
		// clean up first if necessary
		HighBCor.erase(HighBCor.begin(), HighBCor.end());
	}


	// get median for all slices and all segments of the current H2O map
	for (int slice = 1; slice <= nSlices; slice++) {
		H2OIter = H20Map.begin();
		std::advance(H2OIter, slice - 1);
		WMIter = WMMask.begin();
		std::advance(WMIter, slice - 1);
		GMIter = GMMask.begin();
		std::advance(GMIter, slice - 1);
		CSFIter = CSFMask.begin();
		std::advance(CSFIter, slice - 1);

		// copy native H2O map (image) and create H2O map for WM, GM and CSF
		// furthermore, fill vectors with median grey values
		curImg = boost::shared_ptr<qImage>(new qImage(**H2OIter));
		curImg->applyMask(**WMIter, qImage::LARGER_THAN, 0.0);
		H2OWM.push_back(curImg);
		medVecWM.push_back( curImg->getMedianRestriced(qImage::LARGER_THAN, 0) );

		curImg = boost::shared_ptr<qImage>(new qImage(**H2OIter));
		curImg->applyMask(**GMIter, qImage::LARGER_THAN, 0.0);
		H2OGM.push_back(curImg);
		medVecGM.push_back( curImg->getMedianRestriced(qImage::LARGER_THAN, 0) );

		curImg = boost::shared_ptr<qImage>(new qImage(**H2OIter));
		curImg->applyMask(**CSFIter, qImage::LARGER_THAN, 0.0);
		H2OCSF.push_back(curImg);
		medVecCSF.push_back( curImg->getMedianRestriced(qImage::LARGER_THAN, 0) );
	}

	//std::cout << "H2O Median (WM)  = "<<medianWM<<std::endl;
	//std::cout << "H2O Median (GM)  = "<<medianGM<<std::endl;
	//std::cout << "H2O Median (CSF) = "<<medianCSF<<std::endl;

	// normalize each segment with respect to its median H2O content and add them up to
	// get the (yet unsmoothed) high-B correction map
	for (int slice = 0; slice < nSlices; slice++) {
		WMIter = H2OWM.begin();
		std::advance(WMIter, slice);
		GMIter = H2OGM.begin();
		std::advance(GMIter, slice);
		CSFIter = H2OCSF.begin();
		std::advance(CSFIter, slice);

		(*WMIter)->multiply(1000.0/medianWM);
		(*GMIter)->multiply(1000.0/medianGM);
		(*CSFIter)->multiply(1000.0/medianCSF);

/*
		double normWM = medVecWM[slice] > 0 ? 1000.0/medVecWM[slice] : 1000.0/medianWM;
		double normGM = medVecGM[slice] > 0 ? 1000.0/medVecGM[slice] : 1000.0/medianGM;
		double normCSF = medVecCSF[slice] > 0 ? 1000.0/medVecCSF[slice] : 1000.0/medianCSF;
		(*WMIter)->multiply( normWM );
		(*GMIter)->multiply( normGM );
		(*CSFIter)->multiply( normCSF);
*/

		// add up all segments
		curImg = boost::shared_ptr<qImage>(new qImage(**WMIter));
		curImg->add(**GMIter);
		curImg->add(**CSFIter);

		// set pixels which are currently not included to one (thousand)
//		for(int row=0; row<curImg->getNrows(); row++ )
//			for(int col=0; col<curImg->getNcols(); col++ )
//				if( curImg->getPixel(row,col) == 0 )
//					curImg->setPixel(row,col,1000);

		HighBCor.push_back(curImg);
	}

	// filter stack again with 2D mean filter
	FilterKernel flipMapKern;
	flipMapKern.createMeanKernel(9,9,1); // ... we only need the size of the filter mask for the median filter
	qImageProcessing::filterStackWithKernel(HighBCor,flipMapKern, qImageProcessing::FILL_ZEROS_ONLY, qImageProcessing::KERNEL_FILTER);
	flipMapKern.createMeanKernel(11,11,3); // ... we only need the size of the filter mask for the median filter
	qImageProcessing::filterStackWithKernel(HighBCor,flipMapKern, qImageProcessing::OMIT_ZEROS, qImageProcessing::KERNEL_FILTER);
}



std::list<boost::shared_ptr<qImage> > qH2OMapping::calculateEffectiveM0(
		std::list<boost::shared_ptr<qImage> > M0T2Star,
		std::list<boost::shared_ptr<qImage> > M0Myelin,
		std::list<boost::shared_ptr<qImage> > M0PolyFit,
		std::list<boost::shared_ptr<qImage> > T2StarMap
		) {

//	double diffNormCutOff = 30;
	double diffNormCutOff = 15;
	std::list<boost::shared_ptr<qImage> >::iterator M0T2Iter;
	std::list<boost::shared_ptr<qImage> >::iterator T2StarIter;
	std::list<boost::shared_ptr<qImage> >::iterator M0PolyIter;
	std::list<boost::shared_ptr<qImage> >::iterator M0MyelinIter;
	std::list<boost::shared_ptr<qImage> >::iterator M0effIter;
	std::list<boost::shared_ptr<qImage> >::iterator DiffNormIter;
	boost::shared_ptr<qImage> curImg;

	// we need to determine the image size as it is not a priori clear that this function is called before
	// addRelevantData(). In fact, the inverse should be the regular case.
	int nrow = (*M0T2Star.begin())->getNrows();
	int ncol = (*M0T2Star.begin())->getNcols();
	int nslices = M0T2Star.size();

	std::list<boost::shared_ptr<qImage> > M0eff = qImageUtils::getImageList_DeepCopy<qImageMap>(M0T2Star);
	std::list<boost::shared_ptr<qImage> > DiffNorm = qImageUtils::getImageList_DeepCopy<qImageMap>(M0Myelin);

	// Calculate normalized difference between M0(myelin) and M0(T2*)
	for (int slice = 0; slice < nslices; slice++) {

		// set iterators properly (required for parallel processing)
		M0T2Iter = M0T2Star.begin();
		std::advance(M0T2Iter, slice);
		DiffNormIter = DiffNorm.begin();
		std::advance(DiffNormIter, slice);

		(*DiffNormIter)->substract(**M0T2Iter);      // M0_myelin - M0_T2*
		(*DiffNormIter)->multiply(1000.0);           // (M0_myelin - M0_T2*)*1000;
		(*DiffNormIter)->divide(**M0T2Iter);         // (M0_myelin - M0_T2*) / M0_T2* * 1000;
	}

	// filter normalized difference image
	FilterKernel Kern;
	Kern.createMeanKernel(3,3,3); // ... we only need the size of the filter mask for the median filter
	qImageProcessing::filterStackWithKernel(DiffNorm,Kern, qImageProcessing::NONE, qImageProcessing::MEDIAN_FILTER);


	Kern.createMeanKernel(9,9,3); // ... we only need the size of the filter mask for the median filter

	// Replace M0 from M0(myelin) if necessary
	for (int slice = 0; slice < nslices; slice++) {

		// set iterators properly (required for parallel processing)
		M0MyelinIter = M0Myelin.begin();
		std::advance(M0MyelinIter, slice);
		M0effIter = M0eff.begin();
		std::advance(M0effIter, slice);
		DiffNormIter = DiffNorm.begin();
		std::advance(DiffNormIter, slice);
		T2StarIter = T2StarMap.begin();
		std::advance(T2StarIter, slice);
		M0PolyIter = M0PolyFit.begin();
		std::advance(M0PolyIter, slice);

		// filter myelin image first
		curImg = boost::shared_ptr<qImage>( new qImage(**M0MyelinIter));
		qImageProcessing::filterWithKernel(curImg, Kern, qImageProcessing::OMIT_ZEROS, qImageProcessing::REMOVE_OUTLIER, 0.08 );

		matrix<int>& DiffNormMat = (*DiffNormIter)->getPixelMatrix();
		matrix<int>& M0MyelinMatSmooth = curImg->getPixelMatrix();
		matrix<int>& M0MyelinMat = (*M0MyelinIter)->getPixelMatrix();
		matrix<int>& M0effMat = (*M0effIter)->getPixelMatrix();
		matrix<int>& M0PolyMat = (*M0PolyIter)->getPixelMatrix();
		matrix<int>& T2StarMat = (*T2StarIter)->getPixelMatrix();

		for(int row=0; row<nrow; row++ )
			for(int col=0; col<ncol; col++ ) {
				if( DiffNormMat( row,col ) > diffNormCutOff && T2StarMat(row,col) < 200 ) {
					if( DiffNormMat(row,col) < 0 )
						M0effMat(row,col) = M0MyelinMat(row,col);
					else
						M0effMat(row,col) = M0MyelinMatSmooth(row,col);
				}
				if( T2StarMat(row,col) < 29 )
					M0effMat(row,col) = (int)( 1.08*M0PolyMat(row,col) );
			}
	}


	return M0eff;
}


void qH2OMapping::addRelevantData(std::list<boost::shared_ptr<qImage> > T1Map,
		std::list<boost::shared_ptr<qImage> > MOMap,
		std::list<boost::shared_ptr<qImage> > FlipMapCorrection,
		std::list<boost::shared_ptr<qImage> > ReferenceSignalMask,
		std::list<boost::shared_ptr<qImage> > EpiBody,
		std::list<boost::shared_ptr<qImage> > EpiHead,
		std::list<boost::shared_ptr<qImage> > MaskLowResolution,
		double FlipQuteLow,
		double TRQuteLow,
		int numSlices) {
	boost::shared_ptr<qImageMap> curImg;
	boost::numeric::ublas::vector<int> imageDimVec(3);
	std::list<boost::shared_ptr<qImage> >::iterator imageIter;

	// assign member variables
	T1Maps = T1Map;
	M0 = MOMap;
	FlipMapCor = FlipMapCorrection;
	RefSIMask = ReferenceSignalMask;
	EPIBody = EpiBody;
	EPIHead = EpiHead;
	MaskLowRes = MaskLowResolution;

	FlipAngle = FlipQuteLow;
	TR = TRQuteLow;
	nSlices = numSlices;
	imageIter = T1Map.begin();
	nRow = (*imageIter)->getNrows();
	nCol = (*imageIter)->getNcols();

	// create result image objects if not yet created
	if (H20Map.empty() || (int) H20Map.size() != nSlices) {
		// clean up first if necessary
		H20Map.erase(H20Map.begin(), H20Map.end());
		CoilCor.erase(CoilCor.begin(), CoilCor.end());
	}

	for (int slice = 1; slice <= nSlices; slice++) {
		imageDimVec(2) = slice;
		curImg = boost::shared_ptr<qImageMap>(
				new qImageMap(imageDimVec, nRow, nCol));
		H20Map.push_back(curImg);
	}
}



bool qH2OMapping::run(void) {

	std::list<boost::shared_ptr<qImageBinMask> >::iterator maskIter;
	std::list<boost::shared_ptr<qImage> >::iterator T1Iter;
	std::list<boost::shared_ptr<qImage> >::iterator M0Iter;
	std::list<boost::shared_ptr<qImage> >::iterator H2OIter;
	std::list<boost::shared_ptr<qImage> >::iterator CCIter;
	std::list<boost::shared_ptr<qImage> >::iterator FlipCorIter;
	std::list<boost::shared_ptr<qImage> >::iterator RefSIIter;
	std::list<boost::shared_ptr<qImage> >::iterator HighBCorIter;
	std::vector<double> RefSIVec;

	double curT1;
	double curM0;
	double curFlip;
	double curHighBCor;
	double result;
	int slice = 0;

	double defaultT1=3800;

	// consistency check
	if (H20Map.empty()) {
		throw "H2OMapping::run : no data were provided for mapping. Run 'addRelevantData()' first";
	}

	// Check if coil correction has already been calculated
	if( !coilCorCalculated )
		calcCoilCor();

	// Loop over all slices
	for (slice = 1; slice <= nSlices; slice++) {

		// set iterators properly (required for parallel processing)
		H2OIter = H20Map.begin();
		std::advance(H2OIter, slice - 1);
		M0Iter = M0.begin();
		std::advance(M0Iter, slice - 1);
		maskIter = mask.begin();
		std::advance(maskIter, slice - 1);
		T1Iter = T1Maps.begin();
		std::advance(T1Iter, slice - 1);
		CCIter = CoilCor.begin();
		std::advance(CCIter, slice - 1);
		FlipCorIter = FlipMapCor.begin();
		std::advance(FlipCorIter, slice - 1);
		RefSIIter = RefSIMask.begin();
		std::advance(RefSIIter, slice - 1);

		matrix<int>& T1Mat = (*T1Iter)->getPixelMatrix();
		matrix<int>& M0Mat = (*M0Iter)->getPixelMatrix();
		matrix<int>& maskMat = (*maskIter)->getPixelMatrix();
		matrix<int>& CCMat = (*CCIter)->getPixelMatrix();
		matrix<int>& FlipCorMat = (*FlipCorIter)->getPixelMatrix();
		matrix<int>& RefSIMat = (*RefSIIter)->getPixelMatrix();

		// Check if High Field Correction should be applied
		if( applyHighFieldCor ) {
			HighBCorIter = HighBCor.begin();
			std::advance(HighBCorIter, slice - 1);
		}


		// loop over all points in matrix
		for (int row = 0; row < nRow; row++) {
			for (int col = 0; col < nCol; col++) {
				// is current pixel contained in image mask?
				if (maskMat(row, col) == 0) {
					(*H2OIter)->setPixel(row, col, 0.0);
					continue;
				}

				curT1 = T1Mat(row, col);
				curM0 = M0Mat(row, col);
				curFlip = FlipCorMat(row,col)/1000.0;
				if( applyHighFieldCor ) {
					curHighBCor = (*HighBCorIter)->getPixel(row,col)/1000.0;
					curHighBCor = curHighBCor>0 ? 1.0/curHighBCor : 0;
				}
				else
					curHighBCor = 1.0;

				result = curM0 * curHighBCor * (1 - cos(FlipAngle*curFlip) * exp(-TR / curT1))
								/ (1 - exp(-TR / curT1)) * 1 / sin(FlipAngle*curFlip) * CCMat(row,col)/curFlip/1000.0;

				if( RefSIMat(row,col) > 0 ) {
					if( defaultT1 > 0) {
						// Re-calculate signal with default T1
						result = curM0 * curHighBCor * (1 - cos(FlipAngle*curFlip) * exp(-TR / defaultT1))
										/ (1 - exp(-TR / defaultT1)) * 1 / sin(FlipAngle*curFlip) * CCMat(row,col)/curFlip/1000.0;
					}
					RefSIVec.push_back(result);
				}

				(*H2OIter)->setPixel(row, col, (int) (result));
			}
		}
	}

	// determine reference Signal Intensity and normalize image properly
	if( RefSIVec.size() > 0 ) {
		std::sort(RefSIVec.begin(),RefSIVec.end());
		double medianSI = RefSIVec[(int)( RefSIVec.size()/2.0 )];

		// Loop again over all slices to normalize signal intensity
		for (slice = 1; slice <= nSlices; slice++) {

			// set iterators properly (required for parallel processing)
			H2OIter = H20Map.begin();
			std::advance(H2OIter, slice - 1);
			matrix<int>& H2OMat = (*H2OIter)->getPixelMatrix();

			/* loop over all points in matrix */
			for (int row = 0; row < nRow; row++) {
				for (int col = 0; col < nCol; col++) {
					H2OMat(row,col) = (int)(H2OMat(row,col)*1000.0/medianSI);
					H2OMat(row,col) = (int)std::min(H2OMat(row,col),1500);
					H2OMat(row,col) = (int)std::max(H2OMat(row,col),0);
				}
			}
		}
	}

	return true;
}
