//  File name: $HeadURL: https://www.rheinahrcampus.de/svn/repos/predict4a/pLib/src/MyelinMapping.cpp $
//  Revision: $Revision: 54 $
//  Last modified: $Date: 2012-06-01 09:34:24 +0200 (Fr, 01 Jun 2012) $
//  Last modified by: $Author: neeb $

/*
 * MyelinMapping.cpp
 *
 *  Created on: 12.09.2011
 *      Author: neeb

	Copyright 2012 Heiko Neeb

    This file is part of PREDICT-MS (Professional Remagen Diagnostic Image Calculation Tool for MS).

    PREDICT-MS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PREDICT-MS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with "predict".  If not, see <http://www.gnu.org/licenses/>.

	PREDICT-MS is designed to be used for scientific purposes only. Do NOT
	base any diagnosis on the results obtained with this program

	The software is based on the following work which the authors kindly ask to reference in
	any scientific work which is based on PREDICT-MS or parts of it:

	V. Tonkova, V. Arhelger, J. Schenk, H. Neeb. Rapid Myelin Water Content Mapping on Clinical MR Systems.
	Z Med Phys (2012): http://dx.doi.org/10.1016/j.zemedi.2011.09.005

	H. Neeb, V. Ermer, T. Stocker, N.J. Shah. Fast quantitative mapping of absolute water content
	with full brain coverage. NeuroImage (2008). 42(3): 10941109

 */



#include "MyelinMapping.h"
#include "T2StarMapping.h"
#include "qImageMap.h"
#include "qImageUtils.h"
#include "qublasUtils.h"
#include "uquadprog.hpp"
#include <boost/shared_ptr.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <list>
#include <algorithm>
#include <ctime>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/normal_distribution.hpp>


qMyelinMapping::qMyelinMapping() {
	NInnerLoosps=2;
	resFacRows=1;
	resFacCols=1;
}


qMyelinMapping::qMyelinMapping(std::string MyelinMappingMatrixFile) {
	simMatrixLocation = MyelinMappingMatrixFile;
	NInnerLoosps=2;
	resFacRows=1;
	resFacCols=1;
}


qMyelinMapping::~qMyelinMapping() {
	// TODO Auto-generated destructor stub
}


void qMyelinMapping::addRelevantData(std::list<boost::shared_ptr<qImage> > QuteLowFlip,
									std::list<boost::shared_ptr<qImage> > T2StarMap,int numSlices,
		                            float SignalNoiseRatio) {

	boost::shared_ptr<qImageMap> curImg;
	boost::numeric::ublas::vector<int> imageDimVec(3);
	std::list< boost::shared_ptr<qImage> >::iterator imageIter;
	std::list< boost::shared_ptr<qImage> > FirstSliceList;
	int count=0;
	bool found=false;

    // assign member variables
	quteLowFlip=QuteLowFlip;
	T2StarSEMap = T2StarMap;
	nSlices=numSlices;
    nTE=(int)( quteLowFlip.size() / nSlices);
    TE.resize(nTE,false);
    imageIter=QuteLowFlip.begin();
    nRow=(*imageIter)->getNrows();
    nCol=(*imageIter)->getNcols();
    SNR = SignalNoiseRatio;

    // determine data source based on number of echos
    if( nTE == 8 )
    	dataSource = IS_8_TE_DATASET;
    else
       	dataSource = IS_10_TE_DATASET;

    // assign reconstruction parameters
    // These are our standard recon parameters for myelin
	if( dataSource == IS_10_TE_DATASET )
	    wlow=5;
	else
	    wlow=7;

	mT2low=10;
	whigh=0.16;
	lowSmoothScale=100;
	highBoundAsym=1;
	lambda=1.0;
	T2StarMax = 230; //ms


    // get TE vector
	FirstSliceList = qImageUtils::getImagesFromSlice(quteLowFlip, 1);
	if( FirstSliceList.empty() || (int)FirstSliceList.size() != nTE )
		throw "MyelinMapping::addRelevantData : Illegal number of images in the first slice of QuteLowFlip :"+(int)FirstSliceList.size();

	// fill ...
	for(imageIter=FirstSliceList.begin(); imageIter!=FirstSliceList.end(); imageIter++)
		TE(count++)= boost::any_cast<double> ((*imageIter)->getHeader().getHeaderParam("EchoTime", &found));


    // allocate memory for all constant matrices and vectors
	T2StarFull = vector<int>(T2StarMax);
	for(int i=1; i<=T2StarMax; i++)
		T2StarFull(i-1)=i;

	Afull.resize(nTE,T2StarMax,false);
	for(int curTE=0; curTE<nTE; curTE++)
		for(int curT2=0; curT2<T2StarMax; curT2++)
			Afull(curTE,curT2) = exp(-TE(curTE)/T2StarFull(curT2));

	lengthFastRelaxInterval = (int(2*wlow)+1);
	A1=subrange(Afull, 0, nTE, mT2low-wlow-1, mT2low+wlow);

    // create result image objects if not yet created
    if( MyelinMap.empty() || (int)MyelinMap.size() != nSlices ) {

    	// clean up first if necessary
    	MyelinMap.erase(MyelinMap.begin(),MyelinMap.end());
    	T2SslowMap.erase(T2SslowMap.begin(),T2SslowMap.end());
    	T2SfastMap.erase(T2SfastMap.begin(),T2SfastMap.end());
   		M0.erase(M0.begin(), M0.end());

        for( int slice=1; slice<=nSlices; slice++ ) {
        	imageDimVec(2)=slice;
            curImg = boost::shared_ptr<qImageMap>(new qImageMap(imageDimVec, nRow, nCol));
            MyelinMap.push_back(curImg);

            curImg = boost::shared_ptr<qImageMap>(new qImageMap(imageDimVec, nRow, nCol));
            T2SslowMap.push_back(curImg);

            curImg = boost::shared_ptr<qImageMap>(new qImageMap(imageDimVec, nRow, nCol));
            T2SfastMap.push_back(curImg);

            curImg = boost::shared_ptr<qImageMap>(new qImageMap(imageDimVec, nRow, nCol));
            M0.push_back(curImg);
        }
    }
	return;
}


// ToDo: This function is a bit messy and needs to be cleaned up. However, it seems to work as desired....
void qMyelinMapping::estimateMyelin(const matrix<float>& simMatrix, const vector<float> y, float curT2StarSE,
		float& curMyelinEstimate, float& curT2SsEstimate, float& curT2SfEstimate, float& curM0Estimate,
		bool useInitialMyelinEstimate) {

	float curMinDist=100000000.0;
	float newMinDist = 0;
	int curMinRowIdx=-1;
	float putMeanHighT2;
	float T2SsLowerBound=0;
	float T2SsUpperBound=0;
	matrix<float> A,Atmp,A2,A3, idMat, Aopt;
	vector<float> yplus, yminus, ytmp, yopt,  yFit,tmp, tmpT2, unityVec;

	matrix<double> G, CE, CI;
	vector<double> g0, ce0, ci0, x;

	int nrowA, ncolA;
	int lengthSlowRelaxInterval=0;
	int count=0;

	// get simulation matrix row where which best fits to the currently estimated T2*SE
	if(curMyelinEstimate < 0)
		curMyelinEstimate = 0.1;
	if(curMyelinEstimate > 0.36)
		curMyelinEstimate = 0.36;

	for(int crow=0; crow<(int)simMatrix.size1(); crow++) {
		newMinDist = fabs(simMatrix(crow,0)-curT2StarSE);
		if( fabs(simMatrix(crow,3)-curMyelinEstimate) < 0.011 && newMinDist < curMinDist ) {
			curMinDist=newMinDist;
			curMinRowIdx=crow;
		}
	}

	// determine range for the slow relaxing peak
	putMeanHighT2 = simMatrix(curMinRowIdx,2);
	T2SsLowerBound=(int)(putMeanHighT2*(1-whigh*highBoundAsym));
	T2SsUpperBound=(int) (putMeanHighT2*(1+whigh));
	T2SsUpperBound = (T2SsUpperBound > T2StarMax) ? T2StarMax : T2SsUpperBound;
	lengthSlowRelaxInterval = T2SsUpperBound-T2SsLowerBound+1;
    if( T2SsLowerBound==0 && T2SsUpperBound==0 )
    	return;

	A=A1;
	Atmp=subrange(Afull, 0, nTE, T2SsLowerBound-1, T2SsUpperBound);
	qublasUtils::hconcat(A,Atmp);
	nrowA = A.size1();
	ncolA = A.size2();

	// now we have to add the matrices required for smoothing the T2* distribution
	A2=zero_matrix<float>(lengthSlowRelaxInterval-1,ncolA);
	idMat = identity_matrix<float>(lengthSlowRelaxInterval-1);
	idMat=idMat*lambda;
	subrange(A2, 0, lengthSlowRelaxInterval-1, ncolA-lengthSlowRelaxInterval, ncolA-1) = idMat;
	subrange(A2, 0, lengthSlowRelaxInterval-1, ncolA-lengthSlowRelaxInterval+1, ncolA) -= idMat;

	A3=A;
	qublasUtils::vconcat(A3,A2);

	A2=zero_matrix<float>(lengthFastRelaxInterval-1,ncolA);
	idMat = identity_matrix<float>(lengthFastRelaxInterval-1);
	idMat=idMat*lambda*lowSmoothScale;
	subrange(A2, 0, lengthFastRelaxInterval-1, 0, lengthFastRelaxInterval-1) = idMat;
	subrange(A2, 0, lengthFastRelaxInterval-1, 1, lengthFastRelaxInterval)   -= idMat;
	qublasUtils::vconcat(A3,A2);


	if( !useInitialMyelinEstimate  ) {
		Aopt = zero_matrix<float>(2,ncolA);
		for(int i=lengthFastRelaxInterval; i<ncolA; i++) {
			Aopt(1,i)=-1/exp( -pow((count++ - lengthSlowRelaxInterval/2),2)/30);
		}
		for(int i=0; i<lengthFastRelaxInterval; i++)
			Aopt(0,i)=-1;
		idMat = identity_matrix<float>(ncolA);
		qublasUtils::vconcat(Aopt,idMat);

		yopt = vector<float>(2);
		yopt(0) = 0.8*y(0);
		yopt(1) = 2*y(0);
	} else {
		Aopt = zero_matrix<float>(2,ncolA);
		for(int i=lengthFastRelaxInterval; i<ncolA; i++) {
			Aopt(1,i)=-1/exp( -pow((count++ - lengthSlowRelaxInterval/2),2)/30);
		}
		for(int i=0; i<lengthFastRelaxInterval; i++) {
			Aopt(0,i)=-1;
		}
		idMat = identity_matrix<float>(ncolA);
		qublasUtils::vconcat(Aopt,idMat);

		yopt = vector<float>(2);
		yopt(0) = 0.8*curMyelinEstimate*y(0);
		yopt(1) = 2*curMyelinEstimate*y(0);
	}
	ytmp = zero_vector<float>(ncolA);
	qublasUtils::concat(yopt,ytmp);


	yFit=y;
	ytmp=zero_vector<float>(lengthFastRelaxInterval+lengthSlowRelaxInterval-2);
	qublasUtils::concat(yFit,ytmp);

	// get matrices in format required by uquadprog()
	if( lambda > 0) {
		G=prod(trans(A3),A3);
		g0=-prod(yFit, A3);
	} else{
		G=prod(trans(A),A);
		g0=-prod(y, A);
	}

	CE=zero_matrix<float>(g0.size(),1);
	CE(4,0)=-0.001;
	CE(5,0)=0.001;
	ce0=zero_vector<float>(1);
	CI=trans(Aopt);
	ci0=yopt;
	x=zero_vector<float>(g0.size());

	double qpRes=0;
#if 0
	qublasUtils::storeMatrixInTextFile(DebugOutFolder+"/ForMatlab/G.txt", G, 0,0);
#endif

	// Call QP-Solver
	qpRes=solve_quadprog(G, g0,  CE, ce0,  CI, ci0, x);

#if 0
	qublasUtils::storeVectorInTextFile(DebugOutFolder+"/ForMatlab/g0.txt", g0, 0);
	qublasUtils::storeMatrixInTextFile(DebugOutFolder+"/ForMatlab/CE.txt", CE, 0,0);
	qublasUtils::storeVectorInTextFile(DebugOutFolder+"/ForMatlab/ce0.txt", ce0, 0);
	qublasUtils::storeMatrixInTextFile(DebugOutFolder+"/ForMatlab/CI.txt", CI, 0,0);
	qublasUtils::storeVectorInTextFile(DebugOutFolder+"/ForMatlab/ci0.txt", ci0, 0);
	qublasUtils::storeVectorInTextFile(DebugOutFolder+"/ForMatlab/x.txt", x, 0);
#endif

	// Calculate Mylein, M0, T2*f, T2*s
	unityVec=vector<float>(lengthFastRelaxInterval, 1.0);
	tmp=subrange(x, 0, lengthFastRelaxInterval);
	float slowAmpl=inner_prod(tmp, unityVec);
	tmpT2=subrange(T2StarFull, 0, lengthFastRelaxInterval);
	curT2SfEstimate = inner_prod(tmp,tmpT2)/slowAmpl;

	unityVec=vector<float>(x.size()-lengthFastRelaxInterval-1, 1.0);
	tmp=subrange(x, lengthFastRelaxInterval+1, x.size());
	float fastAmpl=inner_prod(tmp, unityVec);
	tmpT2=subrange(T2StarFull, T2SsLowerBound, T2SsLowerBound+lengthSlowRelaxInterval-1);
	curT2SsEstimate = inner_prod(tmp,tmpT2)/fastAmpl;


	if( qpRes != std::numeric_limits<double>::infinity()  ) {
		curMyelinEstimate=slowAmpl/(slowAmpl+fastAmpl);
		curMyelinEstimate = (curMyelinEstimate<0) ? 0 : curMyelinEstimate;
		curMyelinEstimate = (curMyelinEstimate>0.7) ? 0.7 : curMyelinEstimate;
		curM0Estimate = slowAmpl+fastAmpl;
	}
	else {
		if( useInitialMyelinEstimate ) {
			std::cout << "Infeasible! : " << curMyelinEstimate <<"  " << y(0)<<std::endl;
		}
		curMyelinEstimate=-2;
	}

	if( !(curMyelinEstimate == curMyelinEstimate) )  //NaN
		curMyelinEstimate = -1;

	return;
}





bool qMyelinMapping::run(void) {

	std::list< boost::shared_ptr<qImage> >::iterator iter;
	std::list< boost::shared_ptr<qImage> > sliceList;
	std::list< boost::shared_ptr<qImageBinMask> >::iterator maskIter;
	std::list< boost::shared_ptr<qImage> >::iterator MyelinIter;
	std::list< boost::shared_ptr<qImage> >::iterator T2SSEIter;
	std::list< boost::shared_ptr<qImage> >::iterator T2SsIter;
	std::list< boost::shared_ptr<qImage> >::iterator T2SfIter;
	vector<double> y(nTE);

	matrix<int> tmpsimMatrix;
	matrix<float> simMatrix;
	matrix<float> tmpMat;
	int simMatRows=0;
	int simMatCols=0;
	int slice=0;
	int count=0;
	int rep=0;
	float curT2StarSE=0;
	float initialMyelinEstimate=0;
	float curMyelinEstimate=0;
	float curT2SsEstimate=0;
	float curT2SfEstimate=0;
	float curM0Estimate=0;
	float maskCoverage=0;


	// load matrix which contains mapping between T2*_SE and T2*_s
	bool found=qublasUtils::fillMatrixFromTextFile(simMatrixLocation, tmpsimMatrix, simMatRows, simMatCols);
	if( !found )
		throw "MyelinMapping::run() : Couldn't find matrix mapping T2*SE and T2*s. Expected location: "+simMatrixLocation;

	// re-scale and select only colums with T2*f = 10ms
	tmpMat = tmpsimMatrix/1000.0;
	qublasUtils::col_equals(2, 10, tmpMat, simMatrix);


	// define data-set specific initial estimate of myelin water content (to be used for the estimation of T2*s)
	if( dataSource == IS_10_TE_DATASET )
		initialMyelinEstimate = 0.12;
	else
		initialMyelinEstimate = 0.12;

	// Loop over all points
#pragma omp parallel for firstprivate(iter, sliceList,maskIter, MyelinIter,T2SSEIter, T2SsIter, T2SfIter,y,simMatrix, tmpMat, simMatRows, simMatCols, count, rep, curT2StarSE, initialMyelinEstimate,curMyelinEstimate,curT2SsEstimate,curT2SfEstimate, curM0Estimate)
	for( slice=1; slice <= nSlices; slice++) {

//		std::cout << "Myelin processing of slice : "<< slice<<std::endl;

//		if(slice<28 || slice >33)
//		if(slice != 26 )
//			continue;

		// set iterators properly (required for parallel processing)
		MyelinIter = MyelinMap.begin(); std::advance(MyelinIter,slice-1);
		T2SSEIter = T2StarSEMap.begin(); std::advance(T2SSEIter,slice-1);
		T2SsIter = T2SslowMap.begin(); std::advance(T2SsIter,slice-1);
		T2SfIter = T2SfastMap.begin(); std::advance(T2SfIter,slice-1);
		maskIter = mask.begin(); std::advance(maskIter,slice-1);



		// get all images (timepoints) for current slice
		sliceList = qImageUtils::getImagesFromSlice(quteLowFlip, slice);
		matrix<int>& maskMat     = (*maskIter)->getPixelMatrix();
		matrix<int>& T2StarSEMat = (*T2SSEIter)->getPixelMatrix();

		/* loop over all points in matrix */
		for(int row=0; row<nRow; row+=resFacRows) {
//			std::cout << "Slice = " << slice <<"   Row = "<<row<<std::endl;
			for(int col=0; col<nCol; col+=resFacCols) {
//				std::cout << "Slice = " << slice <<"   Row = "<<row<<"   Col="<<col<<std::endl;

				// get average T2* and check if current point is contained in binary mask
				curT2StarSE=0;
				maskCoverage=0;
				for(int resRow=0; resRow<resFacRows; resRow++)
					for(int resCol=0; resCol<resFacCols; resCol++)
					{
						curT2StarSE += T2StarSEMat(row+resRow,col+resCol);
						maskCoverage += maskMat(row+resRow,col+resCol);
					}

				curT2StarSE = curT2StarSE / (resFacRows*resFacCols);
				maskCoverage = maskCoverage / (resFacRows*resFacCols);

				// is current pixel contained in image mask? Is T2*_SE large enough?
				if( maskCoverage<0.5 || curT2StarSE < 20 ) {
					(*MyelinIter)->setPixel(row,col,0);
					(*T2SsIter)->setPixel(row,col,0);
					(*T2SfIter)->setPixel(row,col,0);
					continue;
				}


				// fill y-vector
				count=0;
				y.clear();
				for(iter=sliceList.begin(); iter!=sliceList.end(); iter++) {
					for(int resRow=0; resRow<resFacRows; resRow++)
						for(int resCol=0; resCol<resFacCols; resCol++)
							y(count)+=(*iter)->getPixel(row+resRow,col+resCol);
					y(count)=y(count) / (resFacRows*resFacCols);
					count++;
				}

				curMyelinEstimate = initialMyelinEstimate;

				// iteratively call optimization routine
				for( rep=0; rep<NInnerLoosps; rep++ ) {
					if(rep == 0) { lambda=0.01; lowSmoothScale=1; }
					else  { lambda=1; lowSmoothScale=100; }

					try {
						estimateMyelin(simMatrix, y, curT2StarSE, curMyelinEstimate,
								curT2SsEstimate, curT2SfEstimate, curM0Estimate, false);
					}
					catch(...) { // catch all exceptions
						curMyelinEstimate=0;
						curT2SsEstimate=0;
						curT2SfEstimate=0;
						curM0Estimate=0;
						break;
					}
				}

				// store result
				for(int resRow=0; resRow<resFacRows; resRow++)
					for(int resCol=0; resCol<resFacCols; resCol++) {
						(*MyelinIter)->setPixel(row+resRow,col+resCol,(int)(curMyelinEstimate*1000));
						(*T2SsIter)->setPixel(row+resRow,col+resCol,(int)curT2SsEstimate);
						(*T2SfIter)->setPixel(row+resRow,col+resCol,(int)curT2SfEstimate);
					}
			}
		}
	}


	return true;
}




bool qMyelinMapping::determineM0(void) {

	std::list<boost::shared_ptr<qImage> >::iterator iter;
	std::list<boost::shared_ptr<qImage> > sliceList;
	std::list<boost::shared_ptr<qImageBinMask> >::iterator maskIter;
	std::list<boost::shared_ptr<qImage> >::iterator T2SSEIter;
	std::list<boost::shared_ptr<qImage> >::iterator M0Iter;
	vector<double> y(nTE);

	matrix<int> tmpsimMatrix;
	matrix<float> simMatrix;
	matrix<float> tmpMat;
	int simMatRows = 0;
	int simMatCols = 0;
	int slice = 0;
	int count = 0;
	float whighAbs=0;
	float curT2StarSE = 0;
	float initialMyelinEstimate = 0;
	float curMyelinEstimate = 0;
	float curT2SsEstimate = 0;
	float curT2SfEstimate = 0;
	float curM0Estimate = 0;

	// load matrix which contains mapping between T2*_SE and T2*_s
	bool found = qublasUtils::fillMatrixFromTextFile(simMatrixLocation,tmpsimMatrix, simMatRows, simMatCols);
	if (!found) throw "MyelinMapping::run() : Couldn't find matrix mapping T2*SE and T2*s. Expected location: "
				+ simMatrixLocation;

	// re-scale and select only colums with T2*f = 10ms
	tmpMat = tmpsimMatrix / 1000.0;
	qublasUtils::col_equals(2, 10, tmpMat, simMatrix);

	// remember current values
	int cmT2low=mT2low;
	int cwlow=wlow;
	double cwhigh=whigh;
	float clambda=lambda;
	double clowSmoothScale=lowSmoothScale;
	double chighBoundAsym=highBoundAsym;

	// define data-set specific initial estimate of M0
	if (dataSource == IS_10_TE_DATASET)
	    whighAbs=6;
	else
		whighAbs=5.5;

    mT2low=8;
    wlow=5;
	initialMyelinEstimate = 0.12;
	lambda = 0.01;
	lowSmoothScale = 1;

	// Loop over all points
#ifndef IS_WINDOWS
#pragma omp parallel for firstprivate(iter, sliceList,maskIter,T2SSEIter, M0Iter,y,simMatrix, tmpMat, simMatRows, simMatCols, count, curT2StarSE, initialMyelinEstimate,curMyelinEstimate,curT2SsEstimate,curT2SfEstimate, curM0Estimate)
#endif
	for (slice = 1; slice <= nSlices; slice++) {

		//std::cout << "M0 determination of slice : " << slice << std::endl;

		// set iterators properly (required for parallel processing)
		T2SSEIter = T2StarSEMap.begin();
		std::advance(T2SSEIter, slice - 1);
		M0Iter = M0.begin();
		std::advance(M0Iter, slice - 1);
		maskIter = mask.begin();
		std::advance(maskIter, slice - 1);

		// get all images (timepoints) for current slice
		sliceList = qImageUtils::getImagesFromSlice(quteLowFlip, slice);
		matrix<int>& maskMat = (*maskIter)->getPixelMatrix();
		matrix<int>& T2StarSEMat = (*T2SSEIter)->getPixelMatrix();

		/* loop over all points in matrix */
		for (int row = 0; row < nRow; row++) {
	//		std::cout << "Row = "<< row<<std::endl;
			for (int col = 0; col < nCol; col++) {

				curT2StarSE = T2StarSEMat(row, col);

				// adaptive high interval
				whigh=whighAbs/curT2StarSE;

				// is current pixel contained in image mask? Is T2*_SE large enough?
				if (maskMat(row, col) == 0 || curT2StarSE < 20) {
					(*M0Iter)->setPixel(row, col, 0);
					continue;
				}

				// fill y-vector
				count = 0;
				for (iter = sliceList.begin(); iter != sliceList.end(); iter++)
					y(count++) = (*iter)->getPixel(row, col);

				if( curT2StarSE > 3200 ) {
					(*M0Iter)->setPixel(row, col, y(0));
					continue;
				}

				curMyelinEstimate = initialMyelinEstimate;

				// call optimization routine
				try {
					estimateMyelin(simMatrix, y, curT2StarSE, curMyelinEstimate,
							curT2SsEstimate, curT2SfEstimate, curM0Estimate,
							false);
				} catch (...) { // catch all exceptions
					curMyelinEstimate = 0;
					curT2SsEstimate = 0;
					curT2SfEstimate = 0;
					curM0Estimate = 0;
				}

				// store result
				(*M0Iter)->setPixel(row, col, (int) curM0Estimate);
			}
		}
	}


	// reset optimisation values
	mT2low=cmT2low;
	wlow=cwlow;
	whigh=cwhigh;
	lambda=clambda;
	lowSmoothScale=clowSmoothScale;
	highBoundAsym=chighBoundAsym;

	return true;
}




void qMyelinMapping::SimTest(OPTIMISE_WHAT whatToOptimise) {

	std::list< boost::shared_ptr<qImage> >::iterator iter;
	std::list< boost::shared_ptr<qImage> >::iterator MyelinIter;
	std::list< boost::shared_ptr<qImage> >::iterator T2SSEIter;
	std::list< boost::shared_ptr<qImage> >::iterator T2SsIter;
	std::list< boost::shared_ptr<qImage> >::iterator T2SfIter;
	std::list< boost::shared_ptr<qImage> >::iterator M0Iter;
	vector<double> y(nTE);
	boost::random::mt19937 gen;

	matrix<int> tmpsimMatrix;
	matrix<float> simMatrix;
	matrix<float> tmpMat;
	matrix<float> resultMat;
	vector<float> T2SLongV(7);
//	vector<float> T2SLongV(1);
	vector<float> T2SShortV(5);
//	vector<float> T2SShortV(1);
//	vector<float> SNRV(3);
	vector<float> SNRV(1);
	vector<float> FastAmplitudeV(3);
	float meanMyelin=0;
	float meanM0=0;
	float meanM0SE=0;
	int nRepetitions=200;
	int nRuns=0;
	int simMatRows=0;
	int simMatCols=0;
	int slice=0;
	int rep=0;
	int count=0;
	int count1=0;
	float curT2StarSE=0;
	float meanCurT2Star=0;
	float initialMyelinEstimate=0;
	float curMyelinEstimate=0;
	float curT2SsEstimate=0;
	float curT2SfEstimate=0;
	float curM0Estimate=0;
	float curM0SEEstimate=0;
	bool useInitialMyelinEstimate;

	// seed random number generator
	gen.seed(static_cast<unsigned int>(std::time(0)));

	// initialise result matrix
	nRuns=SNRV.size() * FastAmplitudeV.size() * T2SShortV.size() * T2SLongV.size();
	resultMat = matrix<float>(nRuns,8);

	// load matrix which contains mapping between T2*_SE and T2*_s
	bool found=qublasUtils::fillMatrixFromTextFile(simMatrixLocation, tmpsimMatrix, simMatRows, simMatCols);
	if( !found )
		throw "MyelinMapping::run() : Couldn't find matrix mapping T2*SE and T2*s. Expected location: "+simMatrixLocation;

	// re-scale and select only colums with T2*f = 10ms
	tmpMat = tmpsimMatrix/1000.0;
	qublasUtils::col_equals(2, 10, tmpMat, simMatrix);


	// Fill vectors
	T2SLongV(0)=30; T2SLongV(1)=40; T2SLongV(2)=50; T2SLongV(3)=65; T2SLongV(4)=80; T2SLongV(5)=90 ;T2SLongV(6)=140;
	T2SShortV(0)=5; T2SShortV(1)=10; T2SShortV(2)=12; T2SShortV(3)=15; T2SShortV(4)=18;
//	SNRV(0)=40; SNRV(1)=60; SNRV(2)=80;
	SNRV(0)=65;
	FastAmplitudeV(0)=0.04; FastAmplitudeV(1)=0.1; FastAmplitudeV(2)=0.2;
#if 0
	T2SShortV(0)=10;
	T2SLongV(0)=50;
	SNRV(0)=70;
#endif


	slice=30;
	MyelinIter = MyelinMap.begin(); std::advance(MyelinIter,slice-1);
	T2SSEIter = T2StarSEMap.begin(); std::advance(T2SSEIter,slice-1);
	T2SsIter = T2SslowMap.begin(); std::advance(T2SsIter,slice-1);
	T2SfIter = T2SfastMap.begin(); std::advance(T2SfIter,slice-1);
	M0Iter = M0.begin(); std::advance(M0Iter,slice-1);

	// Loop over different SNR
	for( int cSNR = 0; cSNR < (int)SNRV.size(); cSNR++ ){

		SNR=SNRV(cSNR);

		// Loop over different fast amplitudes
		for( int cFastAmpl = 0; cFastAmpl < (int)FastAmplitudeV.size(); cFastAmpl++ ){

			// Loop over different short relaxation times
			for( int cT2S=0; cT2S < (int)T2SShortV.size(); cT2S++ ) {

				// Loop over different long relaxation times
				for( int cT2L=0; cT2L < (int)T2SLongV.size(); cT2L++ ) {

					// repetitions
					count=0;
					for( int cRep=0; cRep < nRepetitions; cRep++ ) {

						// fill y-vector
						boost::random::normal_distribution<float> randn(0,1/SNR);

						// simulate decay curve
						for(int i=0; i<nTE; i++)
							y(i)=1000*(FastAmplitudeV(cFastAmpl)*exp(-TE(i)/T2SShortV(cT2S))+(1-FastAmplitudeV(cFastAmpl))*exp(-TE(i)/T2SLongV(cT2L))+randn(gen));


						// calculate T2*(SE)
						curT2StarSE=qT2StarMapping::calculateT2Star(TE, y, curM0SEEstimate);

						// adaptive parameters
						if( whatToOptimise == OPTIMISE_M0 ) {
							if (dataSource == IS_10_TE_DATASET) {
							    mT2low=8; //15
							    wlow=5; // 5
								whigh=6/curT2StarSE;
								NInnerLoosps = 1;
							}
							else {
							    mT2low=8; //15
							    wlow=5; // 5
								whigh=5.5/curT2StarSE;
								NInnerLoosps = 1;
							}
						}

						curMyelinEstimate = initialMyelinEstimate;

						// iteratively call optimisation routine
						try {
							for( rep=0; rep<NInnerLoosps; rep++ ) {
								if(rep == 0) { lambda=0.01; lowSmoothScale=1; useInitialMyelinEstimate=false;}
								else  { lambda=1; lowSmoothScale=100; useInitialMyelinEstimate=false;}

								curMyelinEstimate = (curMyelinEstimate < 0) ? initialMyelinEstimate : curMyelinEstimate;
								estimateMyelin(simMatrix, y, curT2StarSE, curMyelinEstimate,
										curT2SsEstimate, curT2SfEstimate, curM0Estimate,useInitialMyelinEstimate);
							}
						}
						catch(...) { // catch all exceptions
							continue;
						}

						if( curMyelinEstimate >= 0 ) {
							meanMyelin +=curMyelinEstimate;
							meanCurT2Star += curT2StarSE;
							meanM0 += curM0Estimate;
							meanM0SE += curM0SEEstimate;
							count++;
						}
					}

					std::cout << "Myelin processing for SNR : "<< SNR ;
					std::cout << "  -  a(Fast)  : " << (float)FastAmplitudeV(cFastAmpl) ;
					std::cout << "  -  T2*Short : " << (int)T2SShortV(cT2S) ;
					std::cout << "  -  T2*Long  : " << (int)T2SLongV(cT2L);
					std::cout << "    Myelin : " << meanMyelin/count;
					std::cout << "    T2StarSE : " << meanCurT2Star/count;
					std::cout << "    M0 : " << meanM0/count;
					std::cout << "    M0(SE) : " << meanM0SE/count << std::endl;

					resultMat(count1,0)=SNR;
					resultMat(count1,1)=FastAmplitudeV(cFastAmpl);
					resultMat(count1,2)=T2SShortV(cT2S);
					resultMat(count1,3)=T2SLongV(cT2L);
					resultMat(count1,4)=meanMyelin/count;
					resultMat(count1,5)=meanCurT2Star/count;
					resultMat(count1,6)=meanM0/count;
					resultMat(count1++,7)=meanM0SE/count;
					meanMyelin=0;
					meanCurT2Star=0;
					meanM0 = 0;
					meanM0SE = 0;
				}
			}
		}
	}

	// store result matrix
	if( !qublasUtils::storeMatrixInTextFile("/Users/neeb/Summary/Programming/predict/ResultMat.txt", resultMat, nRuns, 8) )
		throw "MyelinMapping::SimTest() : Could not save result matrix";
}

