//  File name: $HeadURL: https://www.rheinahrcampus.de/svn/repos/predict4a/pLib/src/T2StarMapping.cpp $
//  Revision: $Revision: 53 $
//  Last modified: $Date: 2012-05-02 10:46:15 +0200 (Mi, 02 Mai 2012) $
//  Last modified by: $Author: neeb $

/*
 * T2StarMapping.cpp
 *
 *  Created on: 07.09.2011
 *      Author: neeb

	Copyright 2012 Heiko Neeb

    This file is part of PREDICT-MS (Professional Remagen Diagnostic Image Calculation Tool for MS).

    PREDICT-MS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PREDICT-MS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with "predict".  If not, see <http://www.gnu.org/licenses/>.

	PREDICT-MS is designed to be used for scientific purposes only. Do NOT
	base any diagnosis on the results obtained with this program

	The software is based on the following work which the authors kindly ask to reference in
	any scientific work which is based on PREDICT-MS or parts of it:

	V. Tonkova, V. Arhelger, J. Schenk, H. Neeb. Rapid Myelin Water Content Mapping on Clinical MR Systems.
	Z Med Phys (2012): http://dx.doi.org/10.1016/j.zemedi.2011.09.005

	H. Neeb, V. Ermer, T. Stocker, N.J. Shah. Fast quantitative mapping of absolute water content
	with full brain coverage. NeuroImage (2008). 42(3): 10941109

*/


#include "T2StarMapping.h"
#include "qImageMap.h"
#include "qImageUtils.h"
#include "qublasUtils.h"
#include <boost/shared_ptr.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <list>
#include <algorithm>


qT2StarMapping::qT2StarMapping() {
	// TODO Auto-generated constructor stub

}

qT2StarMapping::~qT2StarMapping() {
	// TODO Auto-generated destructor stub
}


void qT2StarMapping::addRelevantData(std::list<boost::shared_ptr<qImage> > QuteLowFlip, int numSlices) {

    boost::shared_ptr<qImageMap> curImg;
	boost::numeric::ublas::vector<int> imageDimVec(3);
	std::list< boost::shared_ptr<qImage> >::iterator imageIter;
	std::list< boost::shared_ptr<qImage> > FirstSliceList;
	int count=0;
	bool found=false;

    // assign member variables
	quteLowFlip=QuteLowFlip;
	nSlices=numSlices;
    nTE=(int)( quteLowFlip.size() / nSlices);
    TE.resize(nTE,false);
    imageIter=QuteLowFlip.begin();
    nRow=(*imageIter)->getNrows();
    nCol=(*imageIter)->getNcols();

    // get TE vector
	FirstSliceList = qImageUtils::getImagesFromSlice(quteLowFlip, 1);
	if( FirstSliceList.empty() || (int)FirstSliceList.size() != nTE )
		throw "T2StarMapping::addRelevantData : Illegal number of images in the first slice of QuteLowFlip :"+(int)FirstSliceList.size();

	for(imageIter=FirstSliceList.begin(); imageIter!=FirstSliceList.end(); imageIter++)
		TE(count++)= boost::any_cast<double> ((*imageIter)->getHeader().getHeaderParam("EchoTime", &found));

    // create result image objects if not yet created
    if( T2StarMap.empty() || (int)T2StarMap.size() != nSlices ) {
  		// clean up first if necessary
   		T2StarMap.erase(T2StarMap.begin(),T2StarMap.end());
   		M0.erase(M0.begin(), M0.end());

        for( int slice=1; slice<=nSlices; slice++ ) {
        	imageDimVec(2)=slice;
            curImg = boost::shared_ptr<qImageMap>(new qImageMap(imageDimVec, nRow, nCol));
            T2StarMap.push_back(curImg);

            curImg = boost::shared_ptr<qImageMap>(new qImageMap(imageDimVec, nRow, nCol));
            M0.push_back(curImg);
        }
    }
}


bool qT2StarMapping::run(void) {

	std::list< boost::shared_ptr<qImage> >::iterator iter;
	std::list< boost::shared_ptr<qImage> > sliceList;
	std::list< boost::shared_ptr<qImageBinMask> >::iterator maskIter;
	std::list< boost::shared_ptr<qImage> >::iterator T2StarIter;
	std::list< boost::shared_ptr<qImage> >::iterator M0Iter;
	vector<double> y(nTE);
	vector<double> xtmpVec(TE);
	vector<double> ytmpVec(TE);
	vector<double> xytmpVec(TE);
	vector<double> unityVec(TE);
	double xMean=0;
	double yMean=0;
	double xStd=0;
	double yStd=0;
	double xyCorr=0;
	double curT2Star=0;
	int count=0;
	int slice=0;


	// consistency check
	if( T2StarMap.empty() ) {
		throw "T2StarMapping::run : no data were provided for mapping. Run 'addRelevantData()' first";
	}

	// calculate mean and std of TE vector (== "x - vector")
	xMean=norm_1(TE)/nTE;
	xtmpVec=TE-vector<double>(nTE,xMean);
	xStd=inner_prod(xtmpVec,xtmpVec)/nTE;
	unityVec=vector<double>(nTE,1.0);

	// Loop over all slices
//#pragma omp parallel for firstprivate(iter, sliceList,maskIter, T2StarIter,M0Iter,y,xtmpVec,ytmpVec,xytmpVec,unityVec,xMean,yMean,xStd,yStd,xyCorr,curT2Star,count)
	for( slice=1; slice <= nSlices; slice++) {

		// set iterators properly (required for parallel processing)
		T2StarIter = T2StarMap.begin(); std::advance(T2StarIter,slice-1);
		M0Iter = M0.begin(); std::advance(M0Iter,slice-1);
		maskIter = mask.begin(); std::advance(maskIter,slice-1);


		// get all images (timepoints) for current slice
		sliceList = qImageUtils::getImagesFromSlice(quteLowFlip, slice);
		matrix<int>& maskMat = (*maskIter)->getPixelMatrix();

		// loop over all points in matrix
		for(int row=0; row<nRow; row++) {
			for(int col=0; col<nCol; col++) {

				// is current pixel contained in image mask?
				if( maskMat(row,col) == 0) {
					(*T2StarIter)->setPixel(row,col,0.0);
					(*M0Iter)->setPixel(row,col,0.0);
					continue;
				}


				// fill y-vector
				count=0;
				for(iter=sliceList.begin(); iter!=sliceList.end(); iter++) {
					matrix<int>& curMat=(*iter)->getPixelMatrix();
					y(count++)=log(curMat(row,col));
				}

				// calculate T2*
				yMean=inner_prod(y, unityVec)/nTE;
				ytmpVec=y-vector<double>(nTE,yMean);
				yStd=inner_prod(ytmpVec,ytmpVec)/nTE;

				xytmpVec = element_prod(xtmpVec,ytmpVec);     //(x_i-x_bar).*(y_i-y_bar)
				xyCorr = inner_prod(xytmpVec,unityVec)/nTE;

				curT2Star = -xStd/xyCorr;
				curT2Star = (curT2Star > 0) ? curT2Star : 3500.0;
				curT2Star = (curT2Star < 3500) ? curT2Star : 3500.0;


				(*T2StarIter)->setPixel(row,col,(int)(curT2Star));
				(*M0Iter)->setPixel(row,col,(int)exp(yMean+xMean/curT2Star));
			}
		}
	}

	return true;
}


float qT2StarMapping::calculateT2Star(const vector<float> x, const vector<float> yorig, float& M0) {

	vector<float> y=yorig;
	float curT2Star;
	double slope;
	double offset;

	// log-transform y-vector
	for(int i=0; i<(int)x.size(); i++)
		y(i)=log(yorig(i));

	qublasUtils::linearFit(x, y, slope, offset);

	curT2Star = (slope != 0) ? -1.0/slope : 0.0;
	curT2Star = (curT2Star > 0) ? curT2Star : 0.0;
	curT2Star = (curT2Star < 3500) ? curT2Star : 3500.0;
	M0 = (float)exp(offset);

	return curT2Star;
}
