//  File name: $HeadURL: https://www.rheinahrcampus.de/svn/repos/predict4a/pLib/src/qB0Mapping.cpp $
//  Revision: $Revision: 53 $
//  Last modified: $Date: 2012-05-02 10:46:15 +0200 (Mi, 02 Mai 2012) $
//  Last modified by: $Author: neeb $

/*
 * qB0Mapping.cpp
 *
 *  Created on: 17.02.2012
 *      Author: neeb

	Copyright 2012 Heiko Neeb

    This file is part of PREDICT-MS (Professional Remagen Diagnostic Image Calculation Tool for MS).

    PREDICT-MS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PREDICT-MS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with "predict".  If not, see <http://www.gnu.org/licenses/>.

	PREDICT-MS is designed to be used for scientific purposes only. Do NOT
	base any diagnosis on the results obtained with this program

	The software is based on the following work which the authors kindly ask to reference in
	any scientific work which is based on PREDICT-MS or parts of it:

	V. Tonkova, V. Arhelger, J. Schenk, H. Neeb. Rapid Myelin Water Content Mapping on Clinical MR Systems.
	Z Med Phys (2012): http://dx.doi.org/10.1016/j.zemedi.2011.09.005

	H. Neeb, V. Ermer, T. Stocker, N.J. Shah. Fast quantitative mapping of absolute water content
	with full brain coverage. NeuroImage (2008). 42(3): 10941109

 */

#include "qB0Mapping.h"
#include "qImageMap.h"
#include "qImageUtils.h"
#include "qImageProcessing.h"
#include "qublasUtils.h"
#include <boost/shared_ptr.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <list>
#include <algorithm>


qB0Mapping::qB0Mapping() {
	// TODO Auto-generated constructor stub

}

qB0Mapping::~qB0Mapping() {
	// TODO Auto-generated destructor stub
}

void qB0Mapping::addRelevantData(std::list<boost::shared_ptr<qImage> > greLowFlipPhase, int numSlices) {

    boost::shared_ptr<qImageMap> curImg;
	boost::numeric::ublas::vector<int> imageDimVec(3);
	std::list< boost::shared_ptr<qImage> >::iterator imageIter;
	std::list< boost::shared_ptr<qImage> > FirstSliceList;
	int count=0;
	bool found=false;

    // assign member variables
	grePhase=greLowFlipPhase;
	nSlices=numSlices;
    nTE=(int)( grePhase.size() / nSlices);
    TE.resize(nTE,false);
    imageIter=grePhase.begin();
    nRow=(*imageIter)->getNrows();
    nCol=(*imageIter)->getNcols();

    // get TE vector
	FirstSliceList = qImageUtils::getImagesFromSlice(grePhase, 1);
	if( FirstSliceList.empty() || (int)FirstSliceList.size() != nTE )
		throw "B0Mapping::addRelevantData : Illegal number of images in the first slice of greLowFlip_Phase :"+(int)FirstSliceList.size();

	for(imageIter=FirstSliceList.begin(); imageIter!=FirstSliceList.end(); imageIter++)
		TE(count++)= boost::any_cast<double> ((*imageIter)->getHeader().getHeaderParam("EchoTime", &found));


    // create result image objects if not yet created
    if( B0Eff.empty() || (int)B0Eff.size() != nSlices ) {
  		// clean up first if necessary
    	B0Eff.erase(B0Eff.begin(),B0Eff.end());
      	gradB.erase(gradB.begin(),gradB.end());

        for( int slice=1; slice<=nSlices; slice++ ) {
        	imageDimVec(2)=slice;
            curImg = boost::shared_ptr<qImageMap>(new qImageMap(imageDimVec, nRow, nCol));
            B0Eff.push_back(curImg);
            curImg = boost::shared_ptr<qImageMap>(new qImageMap(imageDimVec, nRow, nCol));
            gradB.push_back(curImg);
        }
    }
}



bool qB0Mapping::run(void) {

	std::list< boost::shared_ptr<qImage> >::iterator iter;
	std::list< boost::shared_ptr<qImage> >::iterator filteredIter;
	std::list< boost::shared_ptr<qImage> > sliceList;
	std::list< boost::shared_ptr<qImageBinMask> >::iterator maskIter;
	std::list< boost::shared_ptr<qImage> >::iterator B0Iter;
	std::list< boost::shared_ptr<qImage> >::iterator B0PrevIter;
	std::list< boost::shared_ptr<qImage> >::iterator gradBIter;
	vector<double> y(nTE);
	int count=0;
	int slice=0;
	double phaseJump=0;
	double sign=0;
	double slope=0;
	double offset=0;

	// consistency check
	if( B0Eff.empty() ) {
		throw "qB0Mapping::run : no data were provided for mapping. Run 'addRelevantData()' first";
	}

	// Loop over all slices
	for( slice=1; slice <= nSlices; slice++) {

		// set iterators properly (required for parallel processing)
		B0Iter = B0Eff.begin(); std::advance(B0Iter,slice-1);
		gradBIter = gradB.begin(); std::advance(gradBIter,slice-1);
		maskIter = mask.begin(); std::advance(maskIter,slice-1);
		if( slice > 1 ) {
			B0PrevIter = B0Eff.begin(); std::advance(B0PrevIter,slice-2);
		}


		// get all images (timepoints) for current slice
		sliceList = qImageUtils::getImagesFromSlice(grePhase, slice);
		matrix<int>& maskMat = (*maskIter)->getPixelMatrix();

		// loop over all points in matrix
		for(int row=0; row<nRow; row++) {
			for(int col=0; col<nCol; col++) {

				// is current pixel contained in image mask?
				if( maskMat(row,col) == 0) {
					(*B0Iter)->setPixel(row,col,0.0);
					(*gradBIter)->setPixel(row,col,0.0);
					continue;
				}

				// fill y-vector
				count=0;
				for(iter=sliceList.begin(); iter!=sliceList.end(); iter++) {
					y(count++)= ( (*iter)->getPixel(row,col)-2048 )/2048.0*PI;
				}

				// phase unwrap
				for(int i=1; i<nTE; i++ ) {
					phaseJump=y(i)-y(i-1);
					if( std::abs(phaseJump) > 1.1*PI ) {   // 1.3*pi
						sign=phaseJump/std::abs(phaseJump);
						for(int j=i; j<nTE; j++ ) {
							y(j) -= sign*2*PI;
						}
					}
				}

				// calculate B0eff from linear fit
				qublasUtils::linearFit(TE, y, slope, offset);
				(*B0Iter)->setPixel(row,col,(int)(slope*10000));

				// calculate field gradient
				if( slice > 1 && row>0 && col>0 ) {
					double gx=slope*10000-(*B0Iter)->getPixel(row,col-1);
					double gy=slope*10000-(*B0Iter)->getPixel(row-1,col);
					double gz=(slope*10000 - (*B0PrevIter)->getPixel(row,col))/3;

					(*gradBIter)->setPixel(row,col,(int)( sqrt(gx*gx+gy*gy+gz*gz)) );
				}
			}
		}
	}


	// Filter
	std::list<boost::shared_ptr<qImage> > filteredB0Map = qImageUtils::getImageList_DeepCopy<qImageMap>( B0Eff );
	FilterKernel flipMapKern;
	flipMapKern.createMeanKernel(9,9,3);
	qImageProcessing::filterStackWithKernel(filteredB0Map,flipMapKern, qImageProcessing::NONE, qImageProcessing::MEDIAN_FILTER);

	// Subtract original image from filtered image
	for( slice=1; slice <= nSlices; slice++) {
		B0Iter = B0Eff.begin(); std::advance(B0Iter,slice-1);
		filteredIter = filteredB0Map.begin(); std::advance(filteredIter,slice-1);

		(*B0Iter)->substract(**filteredIter);
	}

	return true;
}

