//  File name: $HeadURL: https://www.rheinahrcampus.de/svn/repos/predict4a/pLib/src/qImage.cpp $
//  Revision: $Revision: 54 $
//  Last modified: $Date: 2012-06-01 09:34:24 +0200 (Fr, 01 Jun 2012) $
//  Last modified by: $Author: neeb $

/*
 * qImage.cpp
 *
 *  Created on: 02.09.2011
 *      Author: neeb

	Copyright 2012 Heiko Neeb

    This file is part of PREDICT-MS (Professional Remagen Diagnostic Image Calculation Tool for MS).

    PREDICT-MS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PREDICT-MS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with "predict".  If not, see <http://www.gnu.org/licenses/>.

	PREDICT-MS is designed to be used for scientific purposes only. Do NOT
	base any diagnosis on the results obtained with this program

 	The software is based on the following work which the authors kindly ask to reference in
	any scientific work which is based on PREDICT-MS or parts of it:

	V. Tonkova, V. Arhelger, J. Schenk, H. Neeb. Rapid Myelin Water Content Mapping on Clinical MR Systems.
	Z Med Phys (2012): http://dx.doi.org/10.1016/j.zemedi.2011.09.005

	H. Neeb, V. Ermer, T. Stocker, N.J. Shah. Fast quantitative mapping of absolute water content
	with full brain coverage. NeuroImage (2008). 42(3): 10941109

 */

#include "qImage.h"
#include <string>
#include <iostream>
#include <fstream>
#include <istream>
#include <sstream>
#include <iterator>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/shared_ptr.hpp>
#include "qublasUtils.h"
#include "qImageHeader.h"
#include "qImageBinMask.h"
#include "gdcmReader.h"
#include "gdcmWriter.h"
#include "gdcmAttribute.h"
#include "gdcmImageReader.h"
#include "gdcmFileDerivation.h"
#include "gdcmUIDGenerator.h"
#include "gdcmImageWriter.h"

using namespace boost::numeric::ublas;


qImage::qImage() {
	nRow = 0;
	nCol = 0;
	pixel = zero_matrix<int>(nRow,nCol);
	dimensionIdx = vector<int>(2);
}

qImage::qImage(int nrow=0, int ncol=0) {
	nRow = nrow;
	nCol = ncol;
	pixel = zero_matrix<int>(nRow,nCol);
	dimensionIdx = vector<int>(2);
}


qImage::qImage(vector<int> dimIdx, int nrow=0, int ncol=0) {
	nRow = nrow;
	nCol = ncol;
	pixel = zero_matrix<int>(nRow,nCol);
	dimensionIdx = dimIdx;
}


qImage::~qImage() {
}


qImage qImage::operator*(double value) {
	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			pixel(row,col) = (int)(pixel(row,col)*value);

	return *this;
}


bool qImage::loadFromTextFile(std::string fName, int &rows, int &cols) {

	bool retval=qublasUtils::fillMatrixFromTextFile(fName, pixel, rows, cols);
	if( retval ) {
		nRow=rows;
		nCol=cols;
		header.setHeaderParam("SOURCE_LOCATION",fName);
    	header.setHeaderParam("EchoTime",4.8+(dimensionIdx(3)-1) * 3.86);
//      header.setHeaderParam("EchoTime",4.8+(imageDimVec(3)-1) * 3.51);
        header.setHeaderParam("RepetitionTime",12212);
        vector<double> imagePositionPatient(3);
        imagePositionPatient(0)=1;
        imagePositionPatient(1)=2;
        imagePositionPatient(2)=dimensionIdx(2);
        header.setHeaderParam("ImagePosPatient",imagePositionPatient);
	}

	return retval;
}


bool qImage::saveImageAsTextFile(std::string fName) {
	return qublasUtils::storeMatrixInTextFile<int>(fName, pixel, nRow, nCol);
}


bool qImage::loadDicom(std::string filename, int &rows, int &cols) {

	// Instanciate the reader:
	gdcm::Reader reader;
	reader.SetFileName(filename.c_str());
	if (!reader.Read()) {
		std::cerr << "Could not read: " << filename << std::endl;
		return false;
	}
	gdcm::File &file = reader.GetFile();
	gdcm::DataSet &ds = file.GetDataSet(); //DataSet aus Header schreiben

	gdcm::Attribute < 0x0018, 0x0080 > at1; //RepetitionTime wird an qImageHeader bergeben
	assert( at1.GetTag() == gdcm::Tag(0x18, 0x80));
	at1.SetFromDataSet(ds);
	gdcm::Attribute<0x0018, 0x0080>::ArrayType v1 = at1.GetValue();
	double rt;
	rt = (double) v1;
	header.setHeaderParam("RepetitionTime", rt);

	gdcm::Attribute < 0x0018, 0x0081 > at2; //EchoTime wird an qImageHeader bergeben
	assert( at2.GetTag() == gdcm::Tag(0x18, 0x81));
	at2.SetFromDataSet(ds);
	gdcm::Attribute<0x0018, 0x0080>::ArrayType v2 = at2.GetValue();
	double et;
	et = (double) v2;
	header.setHeaderParam("EchoTime", et);

	gdcm::Attribute < 0x0018, 0x1314 > at3; // FLip Angle
	assert( at3.GetTag() == gdcm::Tag(0x18, 0x1314));
	at3.SetFromDataSet(ds);
	gdcm::Attribute<0x0018, 0x1314>::ArrayType v3 = at3.GetValue();
	double fa = (double) v3 / 180.0 * PI;
	header.setHeaderParam("FlipAngle", fa);

	gdcm::Attribute < 0x0010, 0x0010 > at4; //Patient Name wird an qImageHeader bergeben
	assert( at4.GetTag() == gdcm::Tag(0x10, 0x10));
	at4.SetFromDataSet(ds);
	gdcm::Attribute<0x0010, 0x0010>::ArrayType v4 = at4.GetValue();
	std::string patName;
	patName = (std::string) v4;
	header.setHeaderParam("PatientName", patName);

	gdcm::Attribute < 0x0010, 0x0040 > at5; //Patient Gender wird an qImageHeader bergeben
	assert( at5.GetTag() == gdcm::Tag(0x10, 0x40));
	at5.SetFromDataSet(ds);
	gdcm::Attribute<0x0010, 0x0040>::ArrayType v5 = at5.GetValue();
	std::string patGender;
	patGender = (std::string) v5;
	header.setHeaderParam("PatientGender", patGender);

	gdcm::Attribute < 0x0010, 0x1010 > at6; //Patient Age wird an qImageHeader bergeben
	assert( at6.GetTag() == gdcm::Tag(0x10, 0x1010));
	at6.SetFromDataSet(ds);
	gdcm::Attribute<0x0010, 0x1010>::ArrayType v6 = at6.GetValue();
	std::string patAge;
	patAge = (std::string) v6;
	header.setHeaderParam("PatientAge", patAge);


	// Instanciate the imagereader:
	gdcm::ImageReader ir;
	ir.SetFileName(filename.c_str());
	if (!ir.Read()) {
		std::cerr << "Could not read: " << filename << std::endl;
		return false;
	}
	const gdcm::Image &gimage = ir.GetImage();
	char *buffer = new char[gimage.GetBufferLength()];
	const unsigned int* dimension = gimage.GetDimensions(); //Imagedimensions
	nRow = (int) dimension[1];
	nCol = (int) dimension[0];
	gimage.GetBuffer(buffer);
	matrix<int> tmpMat(nRow, nCol);

	if (gimage.GetPhotometricInterpretation()
			== gdcm::PhotometricInterpretation::MONOCHROME2) {
		if (gimage.GetPixelFormat() == gdcm::PixelFormat::INT16
				|| gimage.GetPixelFormat() == gdcm::PixelFormat::UINT16) {
			short *buffer16 = (short*) buffer;
			for (int i = 0; i < nRow; i++) {
				for (int j = 0; j < nCol; j++) {
					tmpMat(i, j) = *buffer16;
					buffer16++;
				}
			}
		} else {
			std::cerr << "Pixel Format is: " << gimage.GetPixelFormat()
					<< std::endl;
			delete[] buffer;
			return false;
		}
	} else {
		std::cerr << "Unhandled PhotometricInterpretation: "
				<< gimage.GetPhotometricInterpretation() << std::endl;
		delete[] buffer;
		return false;
	}

	pixel = tmpMat;
	delete[] buffer;

	rows=nRow;
	cols=nCol;

	return true;
}


bool qImage::saveAsDicom(std::string fileName, std::string fileNameOfDefaultHeaderFile) {

	gdcm::Reader reader;
	reader.SetFileName(fileNameOfDefaultHeaderFile.c_str());
	if (!reader.Read()) {
		std::cerr << "Could not read: " << fileNameOfDefaultHeaderFile << std::endl;
		return false;
	}
	gdcm::File &file = reader.GetFile();
	gdcm::DataSet &ds = file.GetDataSet(); //DataSet aus Header schreiben

#if 0
	bool check=true;
	boost::any headerParamRetVal=header.getHeaderParam("RepetitionTime", &check);
	if( check ) {
		double rt=boost::any_cast<double>(headerParamRetVal);
		gdcm::Attribute<0x18, 0x80> rept;
		rept.SetValue(rt);
		ds.ReplaceEmpty(rept.GetAsDataElement());
	}

	check = true;
	headerParamRetVal=header.getHeaderParam("EchoTime", &check);
	if( check ) {
		double et=boost::any_cast<double>(headerParamRetVal);
		gdcm::Attribute<0x0018, 0x0081> echt;
		echt.SetValue(et);
		ds.Replace(echt.GetAsDataElement());
	}

	check = true;
	headerParamRetVal=header.getHeaderParam("FlipAngle", &check);
	if( check ) {
		double fa=boost::any_cast<double>(headerParamRetVal);
		gdcm::Attribute<0x0018, 0x1314> FlipAngle;
		FlipAngle.SetValue(fa);
		ds.Replace(FlipAngle.GetAsDataElement());
	}
#endif

	// Step 1: Fake Image
	gdcm::SmartPointer<gdcm::Image> im = new gdcm::Image;

	char * buffer = new char[nRow * nCol * 2]; // *2, da short 2 Byte, char nur 1 Byte
	int p = 0;
	matrix<int> tmpMat = pixel;

	for (int i = 0; i < nRow; i++) {
		for (int j = 0; j < nCol; j++) {
			buffer[2 * p + 1] = (char) (tmpMat(i, j) >> 8);
			buffer[2 * p] = (char) (tmpMat(i, j) & 0xFF);
			p++;
		}
	}
	unsigned long l = nRow * nCol * 2;

	im->SetNumberOfDimensions(2);
	im->SetDimension(0, nCol);
	im->SetDimension(1, nRow);

	im->GetPixelFormat().SetSamplesPerPixel(1);
	im->SetPhotometricInterpretation(
			gdcm::PhotometricInterpretation::MONOCHROME2);
	im->SetPixelFormat(gdcm::PixelFormat::INT16);

	gdcm::DataElement pixeldata(gdcm::Tag(0x7fe0, 0x0010));
	pixeldata.SetByteValue(buffer, l);
	delete[] buffer;
	im->SetDataElement(pixeldata);

	gdcm::UIDGenerator uid; // helper for uid generation

	gdcm::FileDerivation fd;
	// For the pupose of this execise we will pretend that this image is referencing
	// two source image (we need to generate fake UID for that).
	const char ReferencedSOPClassUID[] = "1.2.840.10008.5.1.4.1.1.7"; // Secondary Capture
	fd.AddReference(ReferencedSOPClassUID, uid.Generate());
	fd.AddReference(ReferencedSOPClassUID, uid.Generate());

	// Again for the purpose of the exercise we will pretend that the image is a
	// multiplanar reformat (MPR):
	// CID 7202 Source Image Purposes of Reference
	// {"DCM",121322,"Source image for image processing operation"},
	fd.SetPurposeOfReferenceCodeSequenceCodeValue(121322);
	// CID 7203 Image Derivation
	// { "DCM",113072,"Multiplanar reformatting" },
	fd.SetDerivationCodeSequenceCodeValue(113072);
	fd.SetFile(file);
	// If all Code Value are ok the filter will execute properly
	if (!fd.Derive()) {
		std::cerr << "Sorry could not derive using input info" << std::endl;
		return false;
	}

	gdcm::ImageWriter w;
	w.SetImage(*im);
	w.SetFile(fd.GetFile());

	// Set the filename:
	w.SetFileName(fileName.c_str());
	if (!w.Write()) {
		std::cout << "could not write" << std::endl;
		return false;
	}

	return true;
}



bool qImage::exportImageFolder( std::string Folder, std::string fileName, std::string FileEnding,
		                       std::list< boost::shared_ptr<qImage> >& imageList,
		                       std::string defaultHeaderFolder ) {

	char sliceNum[4];
	std::list<boost::shared_ptr<qImage> >::iterator imageIter;
	std::list< std::string > dirListNew;
	int nSlices = imageList.size();
	bool retval = true;
	std::string outFile;

    //std::cout << "Exporting " << Folder << " " << std::endl;

    // read directory which contains files from which header information needs to be copied
	std::list< std::string > dirList = qIOUtils::read_directory(defaultHeaderFolder);
	std::list<std::string>::iterator list_iter = dirList.begin();

	// remove any (obvious) none-image file from list
#if 0
	for(std::list<std::string>::iterator list_iter = dirList.begin();
	    list_iter != dirList.end(); list_iter++) {
		if( (*list_iter).rfind(FileEnding) == std::string::npos ) {
			dirList.erase(list_iter);   // does not seem to be a valid file
			list_iter--;
		}
	}
#endif
	list_iter = dirList.begin();
	for(int i=0; i<(int)(dirList.size()); i++) {
		if( (*list_iter).rfind(FileEnding) != std::string::npos ) {
			dirListNew.push_back(*list_iter);  // does not seem to be a valid file
		}
		list_iter++;
	}

	list_iter = dirListNew.begin();

	for(int slice=1; slice<=nSlices; slice++) {
		imageIter = imageList.begin();
		std::advance(imageIter,slice-1);

		sprintf(sliceNum,"%04i",slice);
		outFile = Folder + fileName + sliceNum + FileEnding;

		if( FileEnding.compare(".txt") == 0 ) {
			if( !(retval = (*imageIter)->saveImageAsTextFile(outFile)) )
				return retval;
		} else if( FileEnding.compare(".ima") == 0 || FileEnding.compare(".dcm") == 0 ) {
	    	std::string defaultHeaderfileName=defaultHeaderFolder+*list_iter;
	    	list_iter++;
	    	if( !(retval = (*imageIter)->saveAsDicom(outFile,defaultHeaderfileName)) )
	    		return retval;
		}
	}

	return true;
}


void qImage::applyMask(qImage &mask, MASK_OPTIONS options, double value) {

	matrix<int>& maskMat = mask.getPixelMatrix();

	if( options == EQUALS )	{
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( maskMat(row,col) != value )
					pixel(row,col) = 0;
	}
	else if ( options == LARGER_THAN ) {
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( maskMat(row,col) <= value )
					pixel(row,col) = 0;
	}
	else if ( options == LESS_THAN ) {
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( maskMat(row,col) >= value )
					pixel(row,col) = 0;
	}
	else if ( options == LARGER_EQUAL ) {
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( maskMat(row,col) < value )
					pixel(row,col) = 0;
	}
	else if ( options == LESS_EQUAL ) {
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( maskMat(row,col) > value )
					pixel(row,col) = 0;
	}
}


boost::shared_ptr<qImage> qImage::makeBinary(MASK_OPTIONS options, double cutoff) {

	boost::shared_ptr<qImage> binaryImage = boost::shared_ptr<qImageBinMask>(new qImageBinMask(nRow, nCol));
	matrix<int>& binaryMat = binaryImage->getPixelMatrix();

	if( options == EQUALS )	{
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( pixel(row,col) == cutoff )
					binaryMat(row,col) = 1;
				else
					binaryMat(row,col) = 0;
	}
	else if ( options == LARGER_THAN ) {
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( pixel(row,col) > cutoff )
					binaryMat(row,col) = 1;
				else
					binaryMat(row,col) = 0;
	}
	else if ( options == LESS_THAN ) {
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( pixel(row,col) < cutoff )
					binaryMat(row,col) = 1;
				else
					binaryMat(row,col) = 0;
	}
	else if ( options == LARGER_EQUAL ) {
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( pixel(row,col) >= cutoff )
					binaryMat(row,col) = 1;
				else
					binaryMat(row,col) = 0;
	}
	else if ( options == LESS_EQUAL ) {
		for(int row=0; row<nRow; row++)
			for(int col=0; col<nCol; col++)
				if( pixel(row,col) <= cutoff )
					binaryMat(row,col) = 1;
				else
					binaryMat(row,col) = 0;
	}

	return binaryImage;
}


void qImage::multiply(qImage &second) {
	matrix<int>& secondMat = second.getPixelMatrix();

	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			pixel(row,col) = pixel(row,col)*secondMat(row,col) ;
}


void qImage::multiply(double second) {

	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			pixel(row,col) = (int)(pixel(row,col)*second) ;
}


void qImage::divide(qImage &second) {
	matrix<int>& secondMat = second.getPixelMatrix();

	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			if( secondMat( row,col) != 0)
				pixel(row,col) = (int)( pixel(row,col) / (double) secondMat(row,col) ) ;
			else
				pixel(row,col) = 0;
}


void qImage::add(qImage &second) {
	matrix<int>& secondMat = second.getPixelMatrix();

	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			pixel(row,col) += secondMat(row,col);
}


void qImage::substract(qImage &second) {
	matrix<int>& secondMat = second.getPixelMatrix();

	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			pixel(row,col) -= secondMat(row,col);
}


void qImage::invert() {
	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			if( pixel(row,col) > 0 )
				pixel(row,col) = (int)(100000.0/pixel(row,col));
}



int qImage::getMax(void) const {

	int maxGW=-1000000000;
	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			if( pixel(row,col) > maxGW )
				maxGW=pixel(row,col);
	return maxGW;
}


int qImage::getMin(void) const {

	int minGW=1000000000;
	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			if( pixel(row,col) < minGW )
				minGW=pixel(row,col);
	return minGW;
}


void qImage::getMinMax(int &minGW, int &maxGW) const {

	minGW=1000000000;
	maxGW=-1000000000;
	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++) {
			if( pixel(row,col) < minGW )
				minGW=pixel(row,col);
			if( pixel(row,col) > maxGW )
				maxGW=pixel(row,col);
		}
}


double qImage::getMeanRestricted(MASK_OPTIONS options, double cutoff) const {

	double mean=0;
	int count=0;
	int curGW=0;

	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++) {
			curGW=pixel(row,col);
			if( (options==EQUALS && cutoff!=curGW) ||
				(options==LARGER_THAN && curGW<=cutoff) ||
				(options==LESS_THAN && curGW>=cutoff) ||
				(options==LARGER_EQUAL && curGW<cutoff) ||
				(options==LESS_EQUAL && curGW>cutoff) )
				continue;

			mean += curGW;
			count++;
		}

	mean = (count>0) ? mean/count : 0;
	return mean;
}


double qImage::getMedianRestriced(MASK_OPTIONS options, double cutoff) const {
	double median=0;
	int count=0;
	int curGW=0;
	std::vector<int> medVec;

	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++) {
			curGW=pixel(row,col);
			if( (options==EQUALS && cutoff!=curGW) ||
				(options==LARGER_THAN && curGW<=cutoff) ||
				(options==LESS_THAN && curGW>=cutoff) ||
				(options==LARGER_EQUAL && curGW<cutoff) ||
				(options==LESS_EQUAL && curGW>cutoff) )
				continue;

			medVec.push_back(curGW);
			count++;
		}

	if( count > 0) {
		std::sort(medVec.begin(),medVec.end());
		median = medVec[(int)( count/2.0 )];
	}
	else
		median=0;

	return median;
}


void qImage::getHistogram(matrix<int> &hist) const {

	// determine size of vector first
	int min=0;
	int max=0;
	getMinMax(min,max);

	hist.resize(max-min+1,2,false);
	for(int i=0; i<(max-min)+1; i++) {
		hist(i,0)=min+i;
		hist(i,1)=0;
	}

	// Fill histogram matrix
	for(int row=0; row<nRow; row++)
		for(int col=0; col<nCol; col++)
			hist(pixel(row,col)-min,1)=hist(pixel(row,col)-min,1)+1;
}


void qImage::setHeader( const qImageHeader &pHeader ) {
	header=pHeader;
}

void qImage::setPixel( int row, int col, int value) {
	pixel(row,col) = value;
}

void qImage::setPixelMatrix( matrix<int> &newMat ) {
	pixel = newMat;
}

bool qImage::checkSize( const qImage &second ) {

	if( second.getNrows() != nRow || second.getNcols() != nCol )
		return false;
	return true;
}

