//  File name: $HeadURL: https://www.rheinahrcampus.de/svn/repos/predict4a/pLib/src/qImage.h $
//  Revision: $Revision: 53 $
//  Last modified: $Date: 2012-05-02 10:46:15 +0200 (Mi, 02 Mai 2012) $
//  Last modified by: $Author: neeb $

/*
 * qImage.h
 *
 *  Created on: 02.09.2011
 *      Author: neeb

	Copyright 2012 Heiko Neeb

    This file is part of PREDICT-MS (Professional Remagen Diagnostic Image Calculation Tool for MS).

    PREDICT-MS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PREDICT-MS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with "predict".  If not, see <http://www.gnu.org/licenses/>.

	PREDICT-MS is designed to be used for scientific purposes only. Do NOT
	base any diagnosis on the results obtained with this program

 	The software is based on the following work which the authors kindly ask to reference in
	any scientific work which is based on PREDICT-MS or parts of it:

	V. Tonkova, V. Arhelger, J. Schenk, H. Neeb. Rapid Myelin Water Content Mapping on Clinical MR Systems.
	Z Med Phys (2012): http://dx.doi.org/10.1016/j.zemedi.2011.09.005

	H. Neeb, V. Ermer, T. Stocker, N.J. Shah. Fast quantitative mapping of absolute water content
	with full brain coverage. NeuroImage (2008). 42(3): 10941109

 */

#ifndef QIMAGE_H_
#define QIMAGE_H_

#include <string>
#include <list>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/shared_ptr.hpp>
#include <qImageHeader.h>
#include "qIOUtils.h"
using namespace boost::numeric::ublas;

static float PI = std::atan(1.0f) * 4.0f;


class qImage {

protected:

	int nRow;
	int nCol;
	vector<int> dimensionIdx;   // e.g. 3: slice; 4: timepoint; 5:....

	qImageHeader header;
	matrix<int> pixel;          // magnitude pixels
	bool checkSize( const qImage &second );

public:

	enum MASK_OPTIONS {
		EQUALS, LARGER_THAN, LESS_THAN, LARGER_EQUAL, LESS_EQUAL, NONE
	};

	qImage();
	qImage(int nrow, int ncol);
	qImage(vector<int> dimensionIndex, int nrow, int ncol);
	virtual ~qImage();

	qImage operator*(double value);

	bool loadFromTextFile(std::string fName, int &rows, int &cols);
	bool saveImageAsTextFile(std::string fName);

	bool loadDicom(std::string filename, int &rows, int &cols);
	bool saveAsDicom(std::string fileName, std::string fileNameOfDefaultHeaderFile);

	void applyMask(qImage &mask, MASK_OPTIONS options, double value);
	boost::shared_ptr<qImage> makeBinary(MASK_OPTIONS options, double cutoff);

	void setHeader( const qImageHeader &pHeader );
	void setPixel( int row, int col, int value);
	void setPixelMatrix( matrix<int> &newMat );

	qImageHeader& getHeader(void) { return header; };
	matrix<int>& getPixelMatrix(void) { return pixel; };
	vector<int>& getDimensionIdx(void) { return dimensionIdx; };
	int getPixel(int row, int col) const {return pixel(row,col); };
	int getNrows(void) const { return nRow; };
	int getNcols(void) const { return nCol; };
	int getMax(void) const;
	int getMin(void) const;
	void getMinMax(int &min, int &max) const;
	void getHistogram(matrix<int> &hist) const;
	double getMeanRestricted(MASK_OPTIONS options, double cutoff) const;
	double getMedianRestriced(MASK_OPTIONS options, double cutoff) const;

	void multiply(qImage &second);
	void multiply(double value);
	void divide(qImage &second);
	void add(qImage &second);
	void substract(qImage &second);

	void invert();

	static bool exportImageFolder( std::string Folder, std::string fileName, std::string FileEnding,
			                       std::list< boost::shared_ptr<qImage> >& imageList,
			                       std::string defaultHeaderFolder);

	template <typename T> static bool importImageFolder( std::string Folder, std::string FileEnding, int nSlices,
								                         std::list< boost::shared_ptr<qImage> >& imageList) {
		boost::numeric::ublas::vector<int> imageDimVec(4);
		int count=0;
		boost::shared_ptr<qImage> curImg;
		int nrow=0;
		int ncol=0;

		// import all files from current folder and create image objects
        //std::cout << "Importing " << Folder << " " << std::endl;
		std::list< std::string > dirList = qIOUtils::read_directory(Folder);

		if( dirList.size() == 0 ) {
			std::cout << "Folder empty/not existent:  "<<Folder  << std::endl;
			return false;
		}

		for(std::list<std::string>::iterator list_iter = dirList.begin();
		    list_iter != dirList.end(); list_iter++) {

	    	// check if we have the correct file ending
	    	std::string fileName=Folder+*list_iter;
	    	size_t found = fileName.rfind(FileEnding);
	    	if( (int)(fileName.length()-found) != (int)FileEnding.length() ) {
	    		continue;    // it seems not to be a valid file type: ignore
	    	}

	    	// load image
	    	++count;
	        imageDimVec(2)=(count-1)%nSlices+1;       // slice index
	        imageDimVec(3)=int((count-1)/nSlices)+1;  // timepoint

	        curImg = boost::shared_ptr<qImage>(new T(imageDimVec, nrow, ncol));
			if( FileEnding.compare(".txt") == 0 ) {
		        if(!(curImg->loadFromTextFile(fileName,nrow,ncol))) {
		        	std::cout << "Can not import file:" << fileName << ". Exiting" << std::endl;
		        	return false;
		        }
			} else if ( FileEnding.compare(".ima") == 0 || FileEnding.compare(".dcm") == 0) {
		        if(!(curImg->loadDicom(fileName,nrow,ncol))) {
		        	std::cout << "Can not import file:" << fileName << ". Exiting" << std::endl;
		        	return false;
		        }
			}

	    	imageList.push_back(curImg);
		}

		return true;
	}
};

#endif /* QIMAGE_H_ */
