//  File name: $HeadURL: https://www.rheinahrcampus.de/svn/repos/predict4a/pLib/src/qImageProcessing.cpp $
//  Revision: $Revision: 53 $
//  Last modified: $Date: 2012-05-02 10:46:15 +0200 (Mi, 02 Mai 2012) $
//  Last modified by: $Author: neeb $

/*
 * qImageProcessing.cpp
 *
 *  Created on: 28.01.2012
 *      Author: neeb

	Copyright 2012 Heiko Neeb

    This file is part of PREDICT-MS (Professional Remagen Diagnostic Image Calculation Tool for MS).

    PREDICT-MS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PREDICT-MS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with "predict".  If not, see <http://www.gnu.org/licenses/>.

	PREDICT-MS is designed to be used for scientific purposes only. Do NOT
	base any diagnosis on the results obtained with this program

 	The software is based on the following work which the authors kindly ask to reference in
	any scientific work which is based on PREDICT-MS or parts of it:

	V. Tonkova, V. Arhelger, J. Schenk, H. Neeb. Rapid Myelin Water Content Mapping on Clinical MR Systems.
	Z Med Phys (2012): http://dx.doi.org/10.1016/j.zemedi.2011.09.005

	H. Neeb, V. Ermer, T. Stocker, N.J. Shah. Fast quantitative mapping of absolute water content
	with full brain coverage. NeuroImage (2008). 42(3): 10941109

 */

#include <list>
#include <vector>
#include "qImage.h"
#include "qImageProcessing.h"
#include "qImageUtils.h"
#include "boost/multi_array.hpp"
#include <boost/numeric/ublas/io.hpp>


qImageProcessing::qImageProcessing() {
	// TODO Auto-generated constructor stub

}

qImageProcessing::~qImageProcessing() {
	// TODO Auto-generated destructor stub
}


void FilterKernel::createMeanKernel(int kernX, int kernY, int kernZ) {

	boost::multi_array<double, 3>::extent_gen extents;
	kernel.resize(extents[kernX][kernY][kernZ]);
	kernNorm=1/(double)(kernX*kernY*kernZ);

	// Assign member variables
	sizeX=kernX;
	sizeY=kernY;
	sizeZ=kernZ;

	for (boost::multi_array<double, 3>::index slice = 0; slice != kernZ; ++slice)
		for (boost::multi_array<double, 3>::index row = 0; row != kernY; ++row)
			for (boost::multi_array<double, 3>::index col = 0; col != kernX; ++col)
				kernel[row][col][slice]=kernNorm;
}


void FilterKernel::createEmptyKernel(int kernX, int kernY, int kernZ) {
	sizeX=kernX;
	sizeY=kernY;
	sizeZ=kernZ;

	boost::multi_array<double, 3>::extent_gen extents;
	kernel.resize(extents[kernX][kernY][kernZ]);
}



double FilterKernel::applyFilter(matrix<int>& Image, int x, int y) {

	double filteredValue=0.0;
	for (boost::multi_array<double, 3>::index row = 0; row != sizeY; ++row)
		for (boost::multi_array<double, 3>::index col = 0; col != sizeX; ++col) {
			filteredValue += Image(row-1,col-1)*kernel[row][col][1];
		}

	return filteredValue;
}


double FilterKernel::applyFilter(boost::multi_array<int, 3>& Image, int x, int y, int z) {

	double filteredValue=0.0;
	for (boost::multi_array<double, 3>::index slice = 0; slice != sizeZ; ++slice)
		for (boost::multi_array<double, 3>::index row = 0; row != sizeY; ++row)
			for (boost::multi_array<double, 3>::index col = 0; col != sizeX; ++col) {
				filteredValue += Image[row][col][slice]*kernel[row][col][slice];
			}

	return filteredValue;
}


void qImageProcessing::sliceAverage(std::list<boost::shared_ptr<qImage> >& original,
		std::list<boost::shared_ptr<qImage> >& copy, FilterKernel &kernel) {

	std::list<boost::shared_ptr<qImage> >::iterator originalIter;
	std::list<boost::shared_ptr<qImage> >::iterator copyIter;

	int kernZ=kernel.getKernSizeZ();
	int nSlices=original.size();

	int leftKernSize=-(int)( (kernZ-1)/2 );
	int rightKernSize=(int)( kernZ/2 );

	// Loop over all slices
	for (int slice = 0; slice < nSlices; slice++) {
		copyIter = copy.begin();
		std::advance(copyIter, slice);

		for(int curKern=leftKernSize; curKern<=rightKernSize; curKern++) {
			if( slice+curKern<0 || slice+curKern>nSlices-1)
				continue;

			originalIter = original.begin();
			std::advance(originalIter, slice+curKern);
			qImage curImg=**originalIter;  // deep copy!
			curImg.multiply(kernel.getKernValue(0, 0, curKern-leftKernSize));
			(*copyIter)->add(curImg);
		}
	}
}



void qImageProcessing::filterStackWithKernel(std::list<boost::shared_ptr<qImage> >& image,
		FilterKernel &kernel, qImageProcessing::FILTER_OPTIONS options, qImageProcessing::FILTER_TYPE filterTpye) {

	int startCol, startRow, startSlice, endCol, endRow, endSlice;
	std::list<boost::shared_ptr<qImage> >::iterator imageIter;
	double curKernValue;

	// make deep copy of the image
	boost::multi_array<int, 3> imageCopy;
	qImageUtils::getImageArray_DeepCopy(image, imageCopy);

	// get image and kernel dimensions
	int nSlices=image.size();
	int nRows=(*image.begin())->getNrows();
	int nCols=(*image.begin())->getNcols();
	int kernX=kernel.getKernSizeX();
	int kernY=kernel.getKernSizeY();
	int kernZ=kernel.getKernSizeZ();

	// Check if kernel size is odd. If not, make it!
	if(kernX%2 > 0) kernX += 1;
	if(kernY%2 > 0) kernY += 1;
	if(kernZ%2 > 0) kernZ += 1;

	// Determine kernel extensions
	int kernSizeX=(int)((kernX-1)/2);
	int kernSizeY=(int)((kernY-1)/2);
	int kernSizeZ=(int)((kernZ-1)/2);

	// Loop over all points
	for (int slice = 0; slice < nSlices; slice++) {

		imageIter = image.begin();
		std::advance(imageIter, slice);

		for(int row=0; row<nRows; row++) {
			for( int col=0; col<nCols; col++) {

				if( options == OMIT_ZEROS && imageCopy[row][col][slice] == 0)
					continue;

				if( options == FILL_ZEROS_ONLY && imageCopy[row][col][slice] != 0)
					continue;

				startCol   = (col-kernSizeX)>=0 ? col-kernSizeX : 0;
				startRow   = (row-kernSizeY)>=0 ? row-kernSizeY : 0;
				startSlice = (slice-kernSizeZ)>=0 ? slice-kernSizeZ : 0;
				endCol     = (col+kernSizeX)<nCols ? col+kernSizeX : nCols-1;
				endRow     = (row+kernSizeY)<nRows ? row+kernSizeY : nRows-1;
				endSlice   = (slice+kernSizeZ)<nSlices ? slice+kernSizeZ : nSlices-1;

				double result=0;

				// Check which filter to apply
				if( filterTpye == KERNEL_FILTER ) {
					double sum=0;
					double kernSum=0;
					// Loop over filter kernel
					for (int z = startSlice; z <= endSlice; z++) {
						for(int y=startRow; y<=endRow; y++) {
							for( int x=startCol; x<=endCol; x++) {

								if( (options == OMIT_ZEROS || options == FILL_ZEROS_ONLY )
										&& imageCopy[y][x][z] == 0)
									continue;

								curKernValue = kernel.getKernValueUnnorm(x-col+kernSizeX, y-row+kernSizeY, z-slice+kernSizeZ);
								kernSum += curKernValue;
								sum += curKernValue*imageCopy[y][x][z];
							}
						}
					}
					if( kernSum > 0)
						result=sum/kernSum;
					else
						result=0;
				}
				else if (filterTpye == MEDIAN_FILTER) {
					std::vector<int> greyValues;
					int GWCount = 0;

					// Loop over filter kernel
					for (int z = startSlice; z <= endSlice; z++) {
						for(int y=startRow; y<=endRow; y++) {
							for( int x=startCol; x<=endCol; x++) {

								if( (options == OMIT_ZEROS || options == FILL_ZEROS_ONLY )
										&& imageCopy[y][x][z] == 0)
									continue;

								greyValues.push_back(imageCopy[y][x][z]);
								GWCount++;
							}
						}
					}

					// sort vector and take median value
					std::sort(greyValues.begin(),greyValues.end());
					result = greyValues[(int)( GWCount/2.0 )];
				}

				// store pixel
				(*imageIter)->setPixel(row, col, (int)result);
			}
		}
	}
}



void qImageProcessing::filterWithKernel(boost::shared_ptr<qImage> & image,
		FilterKernel &kernel, qImageProcessing::FILTER_OPTIONS options, qImageProcessing::FILTER_TYPE filterTpye,
		double outlierCutoff) {

	int startCol, startRow, endCol, endRow;
	double curKernValue;

	// make deep copy of the image
	matrix<int> imageCopy = image->getPixelMatrix();

	// get image and kernel dimensions
	int nRows=image->getNrows();
	int nCols=image->getNcols();
	int kernX=kernel.getKernSizeX();
	int kernY=kernel.getKernSizeY();

	// Check if kernel size is odd. If not, make it!
	if(kernX%2 > 0) kernX += 1;
	if(kernY%2 > 0) kernY += 1;

	// Determine kernel extensions
	int kernSizeX=(int)((kernX-1)/2);
	int kernSizeY=(int)((kernY-1)/2);

	// Loop over all points
	for(int row=0; row<nRows; row++) {
		for( int col=0; col<nCols; col++) {

			if( options == OMIT_ZEROS && imageCopy(row,col) == 0)
				continue;

			if( options == FILL_ZEROS_ONLY && imageCopy(row,col) != 0)
				continue;

			startCol   = (col-kernSizeX)>=0 ? col-kernSizeX : 0;
			startRow   = (row-kernSizeY)>=0 ? row-kernSizeY : 0;
			endCol     = (col+kernSizeX)<nCols ? col+kernSizeX : nCols-1;
			endRow     = (row+kernSizeY)<nRows ? row+kernSizeY : nRows-1;

			double result=0;

			// Check which filter to apply
			if( filterTpye == KERNEL_FILTER ) {
				double sum=0;
				double kernSum=0;
				// Loop over filter kernel
				for(int y=startRow; y<=endRow; y++) {
					for( int x=startCol; x<=endCol; x++) {

						if( (options == OMIT_ZEROS || options == FILL_ZEROS_ONLY )
								&& imageCopy(y,x) == 0)
							continue;

						curKernValue = kernel.getKernValueUnnorm(x-col+kernSizeX, y-row+kernSizeY, 0);
						kernSum += curKernValue;
						sum += curKernValue*imageCopy(y,x);
					}
				}
				if( kernSum > 0)
					result=sum/kernSum;
				else
					result=0;
			}
			else if (filterTpye == MEDIAN_FILTER || filterTpye == REMOVE_OUTLIER) {
				std::vector<int> greyValues;
				int GWCount = 0;

				// Loop over filter kernel
				for(int y=startRow; y<=endRow; y++) {
					for( int x=startCol; x<=endCol; x++) {

						if( (options == OMIT_ZEROS || options == FILL_ZEROS_ONLY )
								&& imageCopy(y,x) == 0)
							continue;

						greyValues.push_back(imageCopy(y,x));
						GWCount++;
					}
				}

				// sort vector and take median value
				std::sort(greyValues.begin(),greyValues.end());
				result = greyValues[(int)( GWCount/2.0 )];

				if( filterTpye == REMOVE_OUTLIER ) {
					if( std::abs(imageCopy(row,col)-result) < outlierCutoff * result)
						result = imageCopy(row,col); // Keep original value
				}
			}

			// store pixel
			image->setPixel(row, col, (int)result);
		}
	}
}



void qImageProcessing::getBrainMask(std::list<boost::shared_ptr<qImage> >& T1Map,
		std::list<boost::shared_ptr<qImage> >& H2OMap,
		std::list<boost::shared_ptr<qImage> >& T2StarMap,
		std::list<boost::shared_ptr<qImage> >& BrainMask,
		bool useH2O) {

	std::list<boost::shared_ptr<qImage> >::iterator T1Iter;
	std::list<boost::shared_ptr<qImage> >::iterator H2OIter;
	std::list<boost::shared_ptr<qImage> >::iterator T2Iter;

	// get image dimensions
	int nSlices=T1Map.size();
	int nRows=(*T1Map.begin())->getNrows();
	int nCols=(*T1Map.begin())->getNcols();

	// Loop over all points and pre-segment
	for (int slice = 0; slice < nSlices; slice++) {

		T1Iter = T1Map.begin();
		std::advance(T1Iter, slice);
		T2Iter = T2StarMap.begin();
		std::advance(T2Iter, slice);
		H2OIter = H2OMap.begin();
		std::advance(H2OIter, slice);

		matrix<int>& T1Mat = (*T1Iter)->getPixelMatrix();
		matrix<int>& T2Mat = (*T2Iter)->getPixelMatrix();
		matrix<int>& H2OMat = (*H2OIter)->getPixelMatrix();

		boost::shared_ptr<qImageBinMask> maskImage = boost::shared_ptr<qImageBinMask>(new qImageBinMask(nRows, nCols));
		matrix<int>& maskMat = maskImage->getPixelMatrix();
		BrainMask.push_back(maskImage);

		if( useH2O )
			for(int row=0; row<nRows; row++) {
				for( int col=0; col<nCols; col++) {
					if( T1Mat(row,col) > 450 && H2OMat(row,col)>550 && T2Mat(row,col)>25 )
						maskMat(row,col)=1;
					else
						maskMat(row,col)=0;
				}
			}
		else
			for(int row=0; row<nRows; row++) {
				for( int col=0; col<nCols; col++) {
					if( T1Mat(row,col) > 450 && T2Mat(row,col)>25 )
						maskMat(row,col)=1;
					else
						maskMat(row,col)=0;
				}
			}

		// Label mask image and get label for the largest connected component
		int labelCount=0;
		int maxLabel=0, maxLabelIdx=0;
		boost::shared_ptr<qImage> labelImg=maskImage->label(labelCount);
		matrix<int> maskHist;
		labelImg->getHistogram(maskHist);
		for( int i=0; i<(int)maskHist.size1(); i++ )
			if( maskHist(i,0)!=0  &&  maskHist(i,1) > maxLabel ) {
				maxLabelIdx=maskHist(i,0);
				maxLabel= maskHist(i,1);
			}


		// remove small connected components in mask
		for(int row=0; row<nRows; row++)
			for( int col=0; col<nCols; col++)
				if( labelImg->getPixel(row,col) != maxLabelIdx )
					maskMat(row,col)=0;


		// finally: fill holes
		StructElem SE;
		SE.createDiskSE(4, 5, 5, 1);
		maskImage->fillHoles(SE);
	}
}


void qImageProcessing::getBinaryMask(std::list<boost::shared_ptr<qImage> >& T1Map,
		std::list<boost::shared_ptr<qImage> >& H2OMap,
		std::list<boost::shared_ptr<qImage> >& T2StarMap,
		std::list<boost::shared_ptr<qImage> >& mask,
		std::list<boost::shared_ptr<qImage> >& superMask,
		int T1low, int T1high, int T2low, int T2high, int H2Olow, int H2Ohigh, bool useSuperMask) {

	std::list<boost::shared_ptr<qImage> >::iterator T1Iter;
	std::list<boost::shared_ptr<qImage> >::iterator H2OIter;
	std::list<boost::shared_ptr<qImage> >::iterator T2Iter;
	std::list<boost::shared_ptr<qImage> >::iterator SuperMaskIter;

	// get image dimensions
	int nSlices=T1Map.size();
	int nRows=(*T1Map.begin())->getNrows();
	int nCols=(*T1Map.begin())->getNcols();

	// delete any possibly existing list entries first
	if ( !mask.empty() ) {
		// clean up first if necessary
		mask.erase(mask.begin(), mask.end());
	}


	// Loop over all points and pre-segment
	for (int slice = 0; slice < nSlices; slice++) {

		T1Iter = T1Map.begin();
		std::advance(T1Iter, slice);
		T2Iter = T2StarMap.begin();
		std::advance(T2Iter, slice);
		H2OIter = H2OMap.begin();
		std::advance(H2OIter, slice);

		matrix<int>& T1Mat = (*T1Iter)->getPixelMatrix();
		matrix<int>& T2Mat = (*T2Iter)->getPixelMatrix();
		matrix<int>& H2OMat = (*H2OIter)->getPixelMatrix();

		boost::shared_ptr<qImageBinMask> maskImage = boost::shared_ptr<qImageBinMask>(new qImageBinMask(nRows, nCols));
		matrix<int>& maskMat = maskImage->getPixelMatrix();
		mask.push_back(maskImage);

		if( useSuperMask ) {
			SuperMaskIter = superMask.begin();
			std::advance(SuperMaskIter, slice);
			matrix<int>& SuperMaskMat = (*SuperMaskIter)->getPixelMatrix();

			for(int row=0; row<nRows; row++) {
				for( int col=0; col<nCols; col++) {
					if( SuperMaskMat(row,col) > 0 &&
							T1Mat(row,col)>=T1low && T1Mat(row,col)<=T1high &&
							T2Mat(row,col)>=T2low && T2Mat(row,col)<=T2high &&
							H2OMat(row,col)>=H2Olow && H2OMat(row,col)<=H2Ohigh )
						maskMat(row,col) = 1;
					else
						maskMat(row,col) = 0;
				}
			}
		}
		else {
			for(int row=0; row<nRows; row++) {
				for( int col=0; col<nCols; col++) {
					if( T1Mat(row,col)>=T1low && T1Mat(row,col)<=T1high &&
							T2Mat(row,col)>=T2low && T2Mat(row,col)<=T2high &&
							H2OMat(row,col)>=H2Olow && H2OMat(row,col)<=H2Ohigh )
						maskMat(row,col) = 1;
					else
						maskMat(row,col) = 0;
				}
			}
		}
	}
}


void qImageProcessing::correctGRELowFlipArtifact(std::list<boost::shared_ptr<qImage> >& GRELowFlip,
		std::list<boost::shared_ptr<qImage> >& T2StarMap,
		std::list<boost::shared_ptr<qImage> >& mask) {

	std::list<boost::shared_ptr<qImage> >::iterator T2Iter;
	std::list<boost::shared_ptr<qImage> >::iterator maskIter;
	std::list<boost::shared_ptr<qImage> >::iterator T2IterNext;
	std::list<boost::shared_ptr<qImage> >::iterator maskIterNext;
	std::list< boost::shared_ptr<qImage> >::iterator iter;
	double weightFirst;

	double T2MinCutoff = 0; // [ms]
	double weighLargerT2Data = 0.2;

	// get image dimensions
	int nSlices=T2StarMap.size();
	int nRows=(*T2StarMap.begin())->getNrows();
	int nCols=(*T2StarMap.begin())->getNcols();

	// temporary copy GRE lowflip data
	std::list< boost::shared_ptr<qImage> > GREcopy=qImageUtils::getImageList_DeepCopy<qImage>(GRELowFlip);

	// Loop over all points
	for (int slice = 0; slice < nSlices-1; slice++) {

		T2Iter = T2StarMap.begin();
		std::advance(T2Iter, slice);
		maskIter = mask.begin();
		std::advance(maskIter, slice);
		T2IterNext = T2StarMap.begin();
		std::advance(T2IterNext, slice+1);
		maskIterNext = mask.begin();
		std::advance(maskIterNext, slice+1);

		matrix<int>& T2Mat = (*T2Iter)->getPixelMatrix();
		matrix<int>& maskMat = (*maskIter)->getPixelMatrix();
		matrix<int>& T2MatNext = (*T2IterNext)->getPixelMatrix();
		matrix<int>& maskMatNext = (*maskIterNext)->getPixelMatrix();

		// get all GRE images (timepoints) for current slice
		std::list< boost::shared_ptr<qImage> > sliceListOrig = qImageUtils::getImagesFromSlice(GRELowFlip, slice);
		std::list< boost::shared_ptr<qImage> > sliceList = qImageUtils::getImagesFromSlice(GREcopy, slice);
		std::list< boost::shared_ptr<qImage> > sliceListNext = qImageUtils::getImagesFromSlice(GREcopy, slice+1);

		for( int row=0; row<nRows; row++) {
			for( int col=0; col<nCols; col++) {

				// Is point in mask (for both slices)
				if( maskMat(row,col) == 0 || maskMatNext(row,col)==0 )
					continue;

				// Check if there is a significant difference in T2* between both slices
				if( abs(T2Mat(row,col)-T2MatNext(row,col)) < T2MinCutoff )
					continue;

				if( T2Mat(row,col) > T2MatNext(row,col) )
					weightFirst=weighLargerT2Data;
				else
					weightFirst=1-weighLargerT2Data;


				// replace GRE lowflip data
				std::list< boost::shared_ptr<qImage> >::iterator iterOrig=sliceListOrig.begin();
				std::list< boost::shared_ptr<qImage> >::iterator iterNext=sliceListNext.begin();
				for(iter=sliceList.begin(); iter!=sliceList.end(); iter++, iterOrig++, iterNext++) {
					matrix<int>& curMat=(*iter)->getPixelMatrix();
					matrix<int>& nextMat=(*iterNext)->getPixelMatrix();
					(*iterOrig)->setPixel( row,col,
							(int) (curMat(row,col)*weightFirst + nextMat(row,col)*(1-weightFirst)) );
				}
			}
		}
	}
}



 void qImageProcessing::motionCorrection(std::list<boost::shared_ptr<qImage> >& magnitude,
		std::list<boost::shared_ptr<qImage> >& phase,
		std::list<boost::shared_ptr<qImage> >& mask,
		PHASE_ENCODE_DIRECTION peDirection,
		int nEchoTimes) {

	 // 1. determine shift value for each phase encode line
	 // -- if nEchotTimes>1
	 // 2. apply shift to magnitude AND phase data
 }



