//  File name: $HeadURL: https://www.rheinahrcampus.de/svn/repos/predict4a/pLib/src/qublasUtils.cpp $
//  Revision: $Revision: 54 $
//  Last modified: $Date: 2012-06-01 09:34:24 +0200 (Fr, 01 Jun 2012) $
//  Last modified by: $Author: neeb $

/*
 * qublasUtils.cpp
 *
 *  Created on: 05.09.2011
 *      Author: neeb

	Copyright 2012 Heiko Neeb

    This file is part of PREDICT-MS (Professional Remagen Diagnostic Image Calculation Tool for MS).

    PREDICT-MS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PREDICT-MS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with "predict".  If not, see <http://www.gnu.org/licenses/>.

	PREDICT-MS is designed to be used for scientific purposes only. Do NOT
	base any diagnosis on the results obtained with this program

 	The software is based on the following work which the authors kindly ask to reference in
	any scientific work which is based on PREDICT-MS or parts of it:

	V. Tonkova, V. Arhelger, J. Schenk, H. Neeb. Rapid Myelin Water Content Mapping on Clinical MR Systems.
	Z Med Phys (2012): http://dx.doi.org/10.1016/j.zemedi.2011.09.005

	H. Neeb, V. Ermer, T. Stocker, N.J. Shah. Fast quantitative mapping of absolute water content
	with full brain coverage. NeuroImage (2008). 42(3): 10941109

 */

#include "qublasUtils.h"
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <string>
#include <iostream>
#include <fstream>
#include <istream>
#include <sstream>
#include <iterator>


using namespace boost::numeric::ublas;

qublasUtils::qublasUtils() {
	// TODO Auto-generated constructor stub

}

qublasUtils::~qublasUtils() {
	// TODO Auto-generated destructor stub
}


void qublasUtils::linearFit(const vector<float> x, const vector<float> y, double &slope, double &offset) {

	int length=x.size();
	vector<double> xtmpVec(length);
	vector<double> ytmpVec(length);
	vector<double> xytmpVec(length);
	vector<double> unityVec(length);
	double xMean=0;
	double yMean=0;
	double xStd=0;
	double yStd=0;
	double xyCorr=0;

	// calculate T2*
	xMean=norm_1(x)/length;
	xtmpVec=x-vector<double>(length,xMean);
	xStd=inner_prod(xtmpVec,xtmpVec)/length;
	unityVec=vector<double>(length,1.0);

	yMean=inner_prod(y, unityVec)/length;
	ytmpVec=y-vector<double>(length,yMean);
	yStd=inner_prod(ytmpVec,ytmpVec)/length;

	xytmpVec = element_prod(xtmpVec,ytmpVec);     //(x_i-x_bar).*(y_i-y_bar)
	xyCorr = inner_prod(xytmpVec,unityVec)/length;

	slope = xStd>0 ? xyCorr/xStd : 1000000000;
	offset = yMean - slope*xMean;
}



bool qublasUtils::fillMatrixFromTextFile(std::string fName, matrix<int> &mat, int &rows, int &cols) {

	std::string Line;
	int curRow = 0;
	int curCol = 0;

	// open file
	const char* const fNameC=fName.c_str();
	std::ifstream in(fNameC);
	if( !in ) {
		std::cout << "can't open file: " << fName << std::endl;
		return false;
	}

	// check if number of rows and columns are already known
	if( rows == 0 || cols == 0 ) {
		using namespace std;

		// read all lines in current document and count
		while( getline( in, Line) ) {

			if( rows == 0 ) {
				istringstream iss (Line, istringstream::in);
				istream_iterator<int> iter(iss);
				for(; iter!=istream_iterator<int>(); iter++) {
					cols++;
				}
			}
			rows++;
		}

		// re-size matrix
		mat.resize(rows,cols,false);

		in.close();
		in.open(fName.c_str());
	}

	// read all lines in current document and fill matrix
	while( getline( in, Line) ) {
		using namespace std;
		istringstream iss (Line, istringstream::in);
		istream_iterator<int> iter(iss);
		for(; iter!=istream_iterator<int>(); iter++) {
			mat(curRow,curCol++)=*iter;
		}
		curRow++;
		curCol=0;
	}


	// close file again
	in.close();

	return true;
}


void qublasUtils::FFT2D(qImage fAbs, qImage fPhase, qImage& FAbs, qImage& FPhase, bool inverse) {

	double curAbs=0;
	double curPhase=0;
	double curReal=0;
	double curImag=0;

	// get original image size
	int nCol = fAbs.getNcols();
	int nRow = fAbs.getNrows();

	// calculate size required for FFT ( multiple of 2^n)
	int l2n = 0, p = 1;
	while (p < nCol) {
		p *= 2;
		l2n++;
	}
	int l2m = 0;
	p = 1; //l2m will become log_2(m)
	while (p < nRow) {
		p *= 2;
		l2m++;
	}

	// determine full image size requird by FFT
	int nRowFFT = 1 << l2m;
	int nColFFT = 1 << l2n;
	int firstRow=(int)( (nRowFFT-nRow)/2 );
	int lastRow=firstRow+nRow-1;
	int firstCol=(int)( (nColFFT-nCol)/2 );
	int lastCol=firstCol+nCol-1;

	// ALlocate memory
	vector<double> mfReal = vector<double>(nRowFFT*nColFFT);
	vector<double> mfImag = vector<double>(nRowFFT*nColFFT);
	vector<double> mFReal = vector<double>(nRowFFT*nColFFT);
	vector<double> mFImag = vector<double>(nRowFFT*nColFFT);


	// Fill original arrays
	for(int col=0; col<nColFFT; col++)
		for(int row=0; row<nRowFFT; row++) {
			if( row >= firstRow && row <=lastRow && col>=firstCol && col<=lastCol ) {
				curAbs = fAbs.getPixel(row-firstRow,col-firstCol);
				curPhase = (fPhase.getPixel(row-firstRow,col-firstCol)-2048)/2048*PI;
				mfReal[nRow * col + row]=curAbs*cos(curPhase);
				mfImag[nRow * col + row]=curAbs*sin(curPhase);

			}
			else {
				mfReal[nRow * col + row]=0;
				mfImag[nRow * col + row]=0;
			}
		}


	// 2D-FFT
	FFT2D(nCol, nRow, inverse, mfReal, mfImag, mFReal, mFImag);

	int x2,y2;
	// Fill FFT image arrays
	for(int col=0; col<nCol; col++)
		for(int row=0; row<nRow; row++) {

			x2 = (col+firstCol + nColFFT / 2) % nColFFT;
			y2 = (row+firstRow + nRowFFT / 2) % nRowFFT;
//			int x2 = col+firstCol;
//			int y2 = row+firstRow;

			curReal=mFReal[nRowFFT * x2 + y2];
			curImag=mFImag[nRowFFT * x2 + y2];

			curReal=mfReal[nRowFFT * x2 + y2];
			curImag=mfImag[nRowFFT * x2 + y2];

			FAbs.setPixel(row,col, (int)(sqrt(curReal*curReal+curImag*curImag)) );
			if( curReal != 0 )
				FPhase.setPixel(row,col, (int)(1000*atan(curImag/curReal)));
			else
				if( curImag < 0 )
					FPhase.setPixel(row,col, (int)(-1000*PI/2));
				else
					FPhase.setPixel(row,col, (int)(+1000*PI/2));
		}
}


// Original version of FFT Code taken from Web-Page http://www.academictutorials.com/graphics/graphics-2d-fourier-transform.asp
// on April, 15th, 2012
void qublasUtils::FFT2D(int n, int m, bool inverse, vector<double>& gRe, vector<double>& gIm,
		vector<double>& GRe, vector<double>& GIm) {
	int l2n = 0, p = 1; //l2n will become log_2(n)
	while (p < n) {
		p *= 2;
		l2n++;
	}
	int l2m = 0;
	p = 1; //l2m will become log_2(m)
	while (p < m) {
		p *= 2;
		l2m++;
	}

	m = 1 << l2m;
	n = 1 << l2n; //Make sure m and n will be powers of 2, otherwise you'll get in an infinite loop

	//Erase all history of this array
	for (int x = 0; x < n; x++) //for each column
		for (int y = 0; y < m; y++) //for each row
			for (int c = 0; c < 1; c++) //for each color component
					{
				GRe[m * x + y ] = gRe[m * x + y ];
				GIm[m * x + y ] = gIm[m * x + y ];
			}

	//Bit reversal of each row
	int j;
	for (int y = 0; y < m; y++) //for each row
		for (int c = 0; c < 1; c++) //for each color component
				{
			j = 0;
			for (int i = 0; i < n - 1; i++) {
				GRe[m * i + y ] = gRe[ m * j + y ];
				GIm[m * i + y ] = gIm[ m * j + y ];
				int k = n / 2;
				while (k <= j) {
					j -= k;
					k /= 2;
				}
				j += k;
			}
		}
	//Bit reversal of each column
	double tx = 0, ty = 0;
	for (int x = 0; x < n; x++) //for each column
		for (int c = 0; c < 1; c++) //for each color component
				{
			j = 0;
			for (int i = 0; i < m - 1; i++) {
				if (i < j) {
					tx = GRe[m * x + i ];
					ty = GIm[m * x + i ];
					GRe[m * x + i ] = GRe[m * x + j ];
					GIm[m * x + i ] = GIm[m * x + j ];
					GRe[m * x + j ] = tx;
					GIm[m * x + j ] = ty;
				}
				int k = m / 2;
				while (k <= j) {
					j -= k;
					k /= 2;
				}
				j += k;
			}
		}

	//Calculate the FFT of the columns
	for (int x = 0; x < n; x++) //for each column
		for (int c = 0; c < 1; c++) //for each color component
				{
			//This is the 1D FFT:
			double ca = -1.0;
			double sa = 0.0;
			int l1 = 1, l2 = 1;
			for (int l = 0; l < l2n; l++) {
				l1 = l2;
				l2 *= 2;
				double u1 = 1.0;
				double u2 = 0.0;
				for (int j = 0; j < l1; j++) {
					for (int i = j; i < n; i += l2) {
						int i1 = i + l1;
						double t1 = u1 * GRe[m * x + i1 ]
								- u2 * GIm[ m * x + i1 ];
						double t2 = u1 * GIm[m * x + i1 ]
								+ u2 * GRe[m * x + i1 ];
						GRe[m * x + i1] = GRe[m * x + i]
								- t1;
						GIm[m * x + i1] = GIm[m * x + i ]
								- t2;
						GRe[m * x + i] += t1;
						GIm[m * x + i] += t2;
					}
					double z = u1 * ca - u2 * sa;
					u2 = u1 * sa + u2 * ca;
					u1 = z;
				}
				sa = sqrt((1.0 - ca) / 2.0);
				if (!inverse)
					sa = -sa;
				ca = sqrt((1.0 + ca) / 2.0);
			}
		}
	//Calculate the FFT of the rows
	for (int y = 0; y < m; y++) //for each row
		for (int c = 0; c < 1; c++) //for each color component
				{
			//This is the 1D FFT:
			double ca = -1.0;
			double sa = 0.0;
			int l1 = 1, l2 = 1;
			for (int l = 0; l < l2m; l++) {
				l1 = l2;
				l2 *= 2;
				double u1 = 1.0;
				double u2 = 0.0;
				for (int j = 0; j < l1; j++) {
					for (int i = j; i < n; i += l2) {
						int i1 = i + l1;
						double t1 = u1 * GRe[m * i1 + y ]
								- u2 * GIm[m * i1 + y ];
						double t2 = u1 * GIm[m * i1 + y ]
								+ u2 * GRe[m * i1 + y ];
						GRe[m * i1 + y] = GRe[m * i + y ]
								- t1;
						GIm[m * i1 + y] = GIm[m * i + y ]
								- t2;
						GRe[m * i +  y] += t1;
						GIm[m * i +  y] += t2;
					}
					double z = u1 * ca - u2 * sa;
					u2 = u1 * sa + u2 * ca;
					u1 = z;
				}
				sa = sqrt((1.0 - ca) / 2.0);
				if (!inverse)
					sa = -sa;
				ca = sqrt((1.0 + ca) / 2.0);
			}
		}

	int d;
	if (inverse)
		d = n;
	else
		d = m;


	for (int x = 0; x < n; x++)
		for (int y = 0; y < m; y++) {
			GRe[m * x + y ] /= d;
			GIm[m * x + y ] /= d;
		}
}


